from __future__ import annotations

import enum
import json
import logging
import uuid
from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    Optional,
    Type,
    TypeVar,
    cast,
)

import asyncpg.pgproto.pgproto
import sqlalchemy as sa
import yarl
from pydantic import BaseModel, TypeAdapter, ValidationError
from sqlalchemy.dialects.postgresql import CIDR, ENUM, JSONB, UUID
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.types import CHAR, SchemaType, TypeDecorator

from ai.backend.appproxy.common.exceptions import InvalidAPIParameters
from ai.backend.appproxy.common.utils import ensure_json_serializable
from ai.backend.common.exception import InvalidIpAddressValue
from ai.backend.common.types import ReadableCIDR
from ai.backend.logging import BraceStyleAdapter

SAFE_MIN_INT = -9007199254740991
SAFE_MAX_INT = 9007199254740991

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

# The common shared metadata instance
convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata_obj = sa.MetaData(naming_convention=convention)
T_StrEnum = TypeVar("T_StrEnum", bound=enum.Enum, covariant=True)


Base = declarative_base(metadata=metadata_obj)


class BaseMixin:
    def dump_model(self, serializable=True) -> dict[str, Any]:
        o = dict(self.__dict__)
        del o["_sa_instance_state"]
        if serializable:
            return ensure_json_serializable(o)
        else:
            return o


pgsql_connect_opts = {
    "server_settings": {
        "jit": "off",
        # 'deadlock_timeout': '10000',  # FIXME: AWS RDS forbids settings this via connection arguments
        "lock_timeout": "60000",  # 60 secs
        "idle_in_transaction_session_timeout": "60000",  # 60 secs
    },
}


# helper functions
def zero_if_none(val):
    return 0 if val is None else val


# FIXME: remove type: ignore annotation as soon as possible
class EnumType(TypeDecorator, SchemaType):  # type: ignore
    """
    A stripped-down version of Spoqa's sqlalchemy-enum34.
    It also handles postgres-specific enum type creation.

    The actual postgres enum choices are taken from the Python enum names.
    """

    impl = ENUM
    cache_ok = True

    def __init__(self, enum_cls, **opts):
        assert issubclass(enum_cls, enum.Enum)
        if "name" not in opts:
            opts["name"] = enum_cls.__name__.lower()
        self._opts = opts
        enums = (m.name for m in enum_cls)
        super().__init__(*enums, **opts)
        self._enum_cls = enum_cls

    def process_bind_param(self, value, dialect):
        return value.name if value else None

    def process_result_value(self, value: Any, dialect):
        return self._enum_cls[value] if value else None

    def copy(self):
        return EnumType(self._enum_cls, **self._opts)

    @property
    def python_type(self):
        return self._enum_class


class StrEnumType(TypeDecorator, Generic[T_StrEnum]):
    """
    Maps Postgres VARCHAR(64) column with a Python enum.StrEnum type.
    """

    impl = sa.VARCHAR
    cache_ok = True

    def __init__(self, enum_cls: type[T_StrEnum], use_name: bool = False, **opts) -> None:
        self._opts = opts
        super().__init__(length=64, **opts)
        self._use_name = use_name
        self._enum_cls = enum_cls

    def process_bind_param(
        self,
        value: Optional[T_StrEnum],
        dialect: sa.Dialect,
    ) -> Optional[str]:
        if value is None:
            return None
        if self._use_name:
            return value.name
        else:
            return value.value

    def process_result_value(
        self,
        value: Optional[str],
        dialect: sa.Dialect,
    ) -> Optional[T_StrEnum]:
        if value is None:
            return None
        if self._use_name:
            return self._enum_cls[value]
        else:
            return self._enum_cls(value)

    def copy(self, **kw):
        return StrEnumType(self._enum_cls, self._use_name, **self._opts)

    @property
    def python_type(self) -> type[T_StrEnum]:
        return self._enum_cls


class StructuredJSONColumn(TypeDecorator):
    """
    A column type to convert JSON values back and forth using a Trafaret.
    """

    impl = JSONB
    cache_ok = True
    _schema: type[BaseModel]

    def __init__(self, schema: type[BaseModel]) -> None:
        super().__init__()
        self._schema = schema

    def load_dialect_impl(self, dialect):
        if dialect.name == "sqlite":
            return dialect.type_descriptor(sa.JSON)
        else:
            return super().load_dialect_impl(dialect)

    def process_bind_param(self, value, dialect):
        if value is None:
            return self._schema()
        try:
            self._schema(**value)
        except ValidationError as e:
            raise ValueError(
                "The given value does not conform with the structured json column format.",
                e.json(),
            )
        return value

    def process_result_value(self, value, dialect):
        if value is None:
            return self._schema()
        return self._schema(**value)

    def copy(self):
        return StructuredJSONColumn(self._schema)


class StructuredJSONObjectColumn(TypeDecorator):
    """
    A column type to convert JSON values back and forth using BaseModel.
    """

    impl = JSONB
    cache_ok = True

    def __init__(self, schema: Type[BaseModel]) -> None:
        super().__init__()
        self._schema = schema

    def process_bind_param(self, value: BaseModel | None, dialect):
        if value:
            return value.model_dump_json()
        return None

    def process_result_value(self, value: str | None, dialect):
        if value:
            return self._schema(**json.loads(value))
        return None

    def copy(self):
        return StructuredJSONObjectColumn(self._schema)


TBaseModel = TypeVar("TBaseModel", bound=BaseModel)


class StructuredJSONObjectListColumn(TypeDecorator, Generic[TBaseModel]):
    """
    A column type to convert JSON values back and forth using BaseModel,
    but store and load a list of the objects.
    """

    impl = JSONB
    cache_ok = True

    def __init__(self, schema: Type[TBaseModel]) -> None:
        super().__init__()
        self._schema = schema

    def process_bind_param(self, value: list[TBaseModel] | None, dialect):
        if value is not None:
            return TypeAdapter(list[TBaseModel]).dump_json(value).decode("utf-8")
        return None

    def process_result_value(self, value: str | None, dialect):
        if value is not None:
            return [self._schema(**i) for i in json.loads(value)]
        return None

    def copy(self):
        return StructuredJSONObjectListColumn(self._schema)


class URLColumn(TypeDecorator):
    """
    A column type for URL strings
    """

    impl = sa.types.UnicodeText
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if isinstance(value, yarl.URL):
            return str(value)
        return value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        if value is not None:
            return yarl.URL(value)


class IPColumn(TypeDecorator):
    """
    A column type to convert IP string values back and forth to CIDR.
    """

    impl = CIDR
    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        try:
            cidr = ReadableCIDR(value).address
        except InvalidIpAddressValue:
            raise InvalidAPIParameters(f"{value} is invalid IP address value")
        return cidr

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return ReadableCIDR(value)


UUID_SubType = TypeVar("UUID_SubType", bound=uuid.UUID)


class GUID(TypeDecorator, Generic[UUID_SubType]):
    """
    Platform-independent GUID type.
    Uses PostgreSQL's UUID type, otherwise uses CHAR(16) storing as raw bytes.
    """

    impl = CHAR
    uuid_subtype_func: ClassVar[Callable[[Any], Any]] = lambda v: v
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(UUID())
        else:
            return dialect.type_descriptor(CHAR(16))

    def process_bind_param(self, value: Any, dialect):
        # NOTE: EndpointId, SessionId, KernelId are *not* actual types defined as classes,
        #       but a "virtual" type that is an identity function at runtime.
        #       The type checker treats them as distinct derivatives of uuid.UUID.
        #       Therefore, we just do isinstance on uuid.UUID only below.
        if value is None:
            return value
        elif dialect.name == "postgresql":
            if isinstance(value, uuid.UUID):
                return str(value)
            else:
                return str(uuid.UUID(value))
        else:
            if isinstance(value, uuid.UUID):
                return value.bytes
            else:
                return uuid.UUID(value).bytes

    def process_result_value(self, value: Any, dialect) -> Optional[UUID_SubType]:
        if value is None:
            return value
        else:
            cls = type(self)
            match value:
                case bytes():
                    return cast(UUID_SubType, cls.uuid_subtype_func(uuid.UUID(bytes=value)))
                case asyncpg.pgproto.pgproto.UUID():
                    return cast(UUID_SubType, cls.uuid_subtype_func(uuid.UUID(str(value))))
                case _:
                    return cast(UUID_SubType, cls.uuid_subtype_func(uuid.UUID(value)))


def IDColumn(name="id"):
    return sa.Column(name, GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()"))


def ForeignKeyIDColumn(name, fk_field, nullable=True):
    return sa.Column(name, GUID, sa.ForeignKey(fk_field), nullable=nullable)



"""add circuit.runtime_variant column

Revision ID: 825d283e0743
Revises: 0dd436073e84
Create Date: 2024-11-11 17:26:43.981090

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "825d283e0743"
down_revision = "0dd436073e84"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("circuits", sa.Column("runtime_variant", sa.VARCHAR(64), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("circuits", "runtime_variant")
    # ### end Alembic commands ###



from typing import TYPE_CHECKING, Sequence
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import relationship, selectinload

from ai.backend.appproxy.common.exceptions import ObjectNotFound
from ai.backend.appproxy.common.types import HealthCheckConfig

from .base import (
    Base,
    BaseMixin,
    IDColumn,
    StructuredJSONObjectColumn,
)

if TYPE_CHECKING:
    pass


__all__ = [
    "Endpoint",
]


class Endpoint(Base, BaseMixin):
    __tablename__ = "endpoints"

    """
    Store model service endpoint information and health check configuration
    """

    id = IDColumn()

    health_check_enabled = sa.Column(sa.Boolean(), nullable=False, default=False)
    health_check_config = sa.Column(StructuredJSONObjectColumn(HealthCheckConfig), nullable=True)

    created_at = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now())
    updated_at = sa.Column(
        sa.DateTime(timezone=True), server_default=sa.func.now(), onupdate=sa.func.now()
    )

    circuit_row = relationship(
        "Circuit",
        back_populates="endpoint_row",
        primaryjoin="Circuit.endpoint_id == Endpoint.id",
        foreign_keys="Circuit.endpoint_id",
        uselist=False,
    )

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        endpoint_id: UUID,
        *,
        load_circuit: bool = False,
    ) -> "Endpoint":
        query = sa.select(Endpoint).where((Endpoint.id == endpoint_id))
        if load_circuit:
            query = query.options(selectinload(Endpoint.circuit_row))
        endpoint = await session.scalar(query)
        if not endpoint:
            raise ObjectNotFound(object_name="endpoint")
        return endpoint

    @classmethod
    async def list_endpoints(
        cls,
        session: AsyncSession,
    ) -> Sequence["Endpoint"]:
        query = sa.select(Endpoint)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def list_health_check_enabled(
        cls,
        session: AsyncSession,
    ) -> Sequence["Endpoint"]:
        query = sa.select(Endpoint).where(Endpoint.health_check_enabled.is_(True))
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    def create(
        cls,
        endpoint_id: UUID,
        health_check_enabled: bool = False,
        health_check_config: HealthCheckConfig | None = None,
    ) -> "Endpoint":
        endpoint = cls()
        endpoint.id = endpoint_id
        endpoint.health_check_enabled = health_check_enabled
        endpoint.health_check_config = health_check_config
        return endpoint



import logging
import uuid
from typing import Iterable

import aiohttp_cors
from aiohttp import web
from pydantic import BaseModel

from ai.backend.appproxy.common.exceptions import AuthorizationFailed
from ai.backend.appproxy.common.types import CORSOptions, PydanticResponse, WebMiddleware
from ai.backend.appproxy.common.utils import pydantic_api_handler
from ai.backend.appproxy.coordinator.api.types import ConfRequestModel
from ai.backend.appproxy.coordinator.types import RootContext
from ai.backend.logging import BraceStyleAdapter

from ..models import Token

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class TokenResponseModel(BaseModel):
    token: str


@pydantic_api_handler(ConfRequestModel)
async def conf_v2(
    request: web.Request, params: ConfRequestModel
) -> PydanticResponse[TokenResponseModel]:
    """
    Generates and returns a token which will be used as an authentication credential for
     /v2/proxy/{token}/{session}/add request.
    """

    root_ctx: RootContext = request.app["_root.context"]
    if not root_ctx.local_config.proxy_coordinator.allow_unauthorized_configure_request:
        token_to_evaluate = request.headers.get("X-BackendAI-Token")
        if token_to_evaluate != root_ctx.local_config.secrets.api_secret:
            raise AuthorizationFailed("Unauthorized access")

    assert params.session.id and params.session.access_key, "Not meant for inference apps"

    async with root_ctx.db.begin_session() as sess:
        token_id = uuid.uuid4()
        token = Token.create(
            token_id,
            params.login_session_token,
            params.kernel_host,
            params.kernel_port,
            params.session.id,
            params.session.user_uuid,
            params.session.group_id,
            params.session.access_key,
            params.session.domain_name,
        )
        sess.add(token)
        log.debug("build token with body {}", params.model_dump())

        return PydanticResponse(TokenResponseModel(token=str(token_id)))


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "v2/conf"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("POST", conf_v2))
    return app, []



from __future__ import annotations

import time
from typing import TYPE_CHECKING, Annotated, Iterable, Sequence
from uuid import UUID

import aiohttp_cors
import sqlalchemy as sa
from aiohttp import web
from pydantic import BaseModel, Field, validator
from sqlalchemy.orm import selectinload

from ai.backend.appproxy.common.types import (
    AppMode,
    CORSOptions,
    PydanticResponse,
    SerializableCircuit,
    WebMiddleware,
)
from ai.backend.appproxy.common.utils import pydantic_api_handler, pydantic_api_response_handler

from ..models import Circuit
from ..models.utils import execute_with_txn_retry
from ..types import RootContext
from .types import StubResponseModel
from .utils import auth_required

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession as SASession


class BulkRemoveCircuitsRequestModel(BaseModel):
    circuit_ids: Annotated[list[UUID], Field(description="Comma separated list of Circuit UUIDs.")]

    @validator("circuit_ids", pre=True)
    def split_comma_separated(cls, v: object) -> object:
        if isinstance(v, str):
            v = v.strip()
            return [] if v == "" else v.split(",")
        return v


class CircuitStatisticsModel(SerializableCircuit):
    requests: int = Field(description="Number of requests processed by this circuit.")
    last_access: int | None = Field(description="Last access timestamp.")
    ttl: int | None = Field(
        description="Number of seconds remaining before this circuit will be discharged due to inactivity. Can be null if `app_mode` is `interactive`.",
    )


@auth_required("worker")
@pydantic_api_response_handler
async def get_circuit(request: web.Request) -> PydanticResponse[SerializableCircuit]:
    """
    Returns information of a circuit.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async with root_ctx.db.begin_readonly_session() as sess:
        circuit = await Circuit.get(sess, UUID(request.match_info["circuit_id"]))
        return PydanticResponse(SerializableCircuit(**circuit.dump_model()))


@auth_required("worker")
@pydantic_api_response_handler
async def delete_circuit(request: web.Request) -> PydanticResponse[StubResponseModel]:
    """
    Removes circuit record from both coordinator and worker.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async def _update(sess: SASession) -> None:
        circuit = await Circuit.get(sess, UUID(request.match_info["circuit_id"]), load_worker=True)
        circuit.worker_row.occupied_slots -= 1
        await sess.delete(circuit)
        await root_ctx.circuit_manager.unload_circuits([circuit])

    async with root_ctx.db.connect() as db_conn:
        await execute_with_txn_retry(_update, root_ctx.db.begin_session, db_conn)
    return PydanticResponse(StubResponseModel(success=True))


@auth_required("worker")
@pydantic_api_handler(BulkRemoveCircuitsRequestModel)
async def delete_circuit_bulk(
    request: web.Request, params: BulkRemoveCircuitsRequestModel
) -> PydanticResponse[StubResponseModel]:
    """
    Removes circuit record from both coordinator and worker, in bulk.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async def _update(sess: SASession) -> Sequence[Circuit]:
        query = (
            sa.select(Circuit)
            .where(Circuit.id.in_(params.circuit_ids))
            .options(selectinload(Circuit.worker_row))
        )
        result = await sess.execute(query)
        circuits = result.scalars().all()

        for circuit in circuits:
            circuit.worker_row.occupied_slots -= 1
            await sess.delete(circuit)

        return circuits

    async with root_ctx.db.connect() as db_conn:
        circuits = await execute_with_txn_retry(_update, root_ctx.db.begin_session, db_conn)
    await root_ctx.circuit_manager.unload_circuits(circuits)

    return PydanticResponse(StubResponseModel(success=True))


@auth_required("worker")
@pydantic_api_response_handler
async def get_circuit_statistics(request: web.Request) -> PydanticResponse[CircuitStatisticsModel]:
    """
    Lists statical informations about given circuit.
    """
    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin_readonly_session() as sess:
        circuit = await Circuit.get(sess, UUID(request.match_info["circuit_id"]))

    last_access, requests = await root_ctx.valkey_live.get_multiple_live_data([
        f"circuit.{circuit.id}.last_access",
        f"circuit.{circuit.id}.requests",
    ])
    # Handle bytes data from valkey
    last_access_value = (
        float(last_access.decode("utf-8")) if last_access else circuit.created_at.timestamp()
    )
    requests_value = int(requests.decode("utf-8")) if requests else 0

    if circuit.app_mode != AppMode.INFERENCE:
        ttl = int(
            root_ctx.local_config.proxy_coordinator.unused_circuit_collection_timeout
            - (time.time() - last_access_value)
        )
    else:
        ttl = None
    return PydanticResponse(
        CircuitStatisticsModel(
            ttl=ttl,
            last_access=int(last_access_value * 1000),
            requests=requests_value,
            **circuit.dump_model(),
        )
    )


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "api/circuit"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    cors.add(add_route("DELETE", "/_/bulk", delete_circuit_bulk))
    cors.add(add_route("GET", "/{circuit_id}", get_circuit))
    cors.add(add_route("GET", "/{circuit_id}/statistics", get_circuit_statistics))
    cors.add(add_route("DELETE", "/{circuit_id}", delete_circuit))
    return app, []



from . import circuit as _circuit
from . import endpoint as _endpoint
from . import token as _token
from . import worker as _worker
from .base import metadata_obj as metadata

__all__ = (
    "metadata",
    *_circuit.__all__,
    *_endpoint.__all__,
    *_worker.__all__,
    *_token.__all__,
)

from .circuit import *  # noqa
from .endpoint import *  # noqa
from .token import *  # noqa
from .worker import *  # noqa



"""add endpoint health check table

Revision ID: a1b2c3d4e5f6
Revises: 7dbbc087108e
Create Date: 2025-06-30 12:00:00.000000

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from ai.backend.appproxy.coordinator.models.base import GUID, convention

# revision identifiers, used by Alembic.
revision = "a1b2c3d4e5f6"
down_revision = "7dbbc087108e"
branch_labels = None
depends_on = None


metadata = sa.MetaData(naming_convention=convention)
circuit = sa.Table(
    "circuits",
    metadata,
    sa.Column("id", GUID, primary_key=True),
    sa.Column("endpoint_id", GUID, primary_key=True),
)
endpoint = sa.Table(
    "endpoints",
    metadata,
    sa.Column("id", GUID, primary_key=True),
)


def upgrade():
    # Create endpoints table
    op.create_table(
        "endpoints",
        sa.Column(
            "id",
            postgresql.UUID(as_uuid=True),
            primary_key=True,
            server_default=sa.text("uuid_generate_v4()"),
        ),
        sa.Column(
            "health_check_enabled",
            sa.Boolean(),
            nullable=False,
            default=False,
            server_default="false",
        ),
        sa.Column("health_check_config", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column("created_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
        sa.Column("updated_at", sa.DateTime(timezone=True), server_default=sa.func.now()),
    )

    conn = op.get_bind()
    query = (
        sa.select(circuit.c.id, circuit.c.endpoint_id)
        .select_from(circuit)
        .where(circuit.c.endpoint_id.is_not(None))
    )
    rows = conn.execute(query).fetchall()
    endpoint_ids = [r.endpoint_id for r in rows]
    query = endpoint.insert().values([{"id": e} for e in endpoint_ids])
    op.execute(query)


def downgrade():
    # Drop endpoints table
    op.drop_table("endpoints")



"""Health monitoring API endpoints"""

import logging
from datetime import datetime
from typing import Iterable, Literal
from uuid import UUID

import aiohttp_cors
import sqlalchemy as sa
from aiohttp import web
from pydantic import BaseModel

from ai.backend.appproxy.common.types import AppMode, CORSOptions, PydanticResponse, WebMiddleware
from ai.backend.appproxy.common.utils import pydantic_api_response_handler
from ai.backend.common.types import ModelServiceStatus
from ai.backend.logging import BraceStyleAdapter

from .. import __version__
from ..models import Circuit, Endpoint, Worker
from ..types import RootContext
from .utils import auth_required

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class RouteHealthStatusModel(BaseModel):
    """Model for individual route health status"""

    route_id: UUID
    session_id: UUID
    kernel_host: str
    kernel_port: int
    protocol: str
    health_status: ModelServiceStatus | None
    last_health_check: float | None
    consecutive_failures: int
    created_at: datetime
    updated_at: datetime


class EndpointHealthStatusModel(BaseModel):
    """Model for endpoint health status summary"""

    endpoint_id: UUID
    health_check_enabled: bool
    total_routes: int
    healthy_routes: int
    unhealthy_routes: int
    unknown_routes: int
    routes: list[RouteHealthStatusModel]


class CircuitHealthStatusModel(BaseModel):
    """Model for circuit health status with route details"""

    circuit_id: UUID
    endpoint_id: UUID | None
    total_routes: int
    healthy_routes: int
    all_routes: list[RouteHealthStatusModel]


class HealthSummaryModel(BaseModel):
    """Model for overall health summary"""

    total_endpoints: int
    endpoints_with_health_checks: int
    total_routes: int
    healthy_routes: int
    unhealthy_routes: int
    unknown_routes: int


class HealthStatusResponseModel(BaseModel):
    """Response model for health status APIs"""

    success: bool
    summary: HealthSummaryModel | None = None
    endpoints: list[EndpointHealthStatusModel] | None = None
    circuits: list[CircuitHealthStatusModel] | None = None


class WorkerInfoModel(BaseModel):
    authority: str
    available_slots: int
    occupied_slots: int
    ha_setup: bool


class StatusResponseModel(BaseModel):
    coordinator_version: str
    appproxy_api_version: Literal["v2"]
    workers: list[WorkerInfoModel]
    """
    List of active workers (pool of workers which are able to handle proxying job).
    """


@auth_required("manager")
@pydantic_api_response_handler
async def get_health_summary(request: web.Request) -> PydanticResponse[HealthStatusResponseModel]:
    """Get overall health status summary"""
    root_ctx: RootContext = request.app["ctx"]

    async with root_ctx.db.begin_readonly_session() as sess:
        # Get endpoint counts
        endpoints_query = sa.select(sa.func.count(Endpoint.id))
        total_endpoints = await sess.scalar(endpoints_query) or 0

        health_enabled_query = sa.select(sa.func.count(Endpoint.id)).where(
            Endpoint.health_check_enabled.is_(True)
        )
        endpoints_with_health = await sess.scalar(health_enabled_query) or 0

        # Get route health counts by examining circuit route_info JSON
        total_routes = 0
        healthy_routes = 0
        unhealthy_routes = 0
        unknown_routes = 0

        # Query all circuits that have endpoints with health checking enabled and are in INFERENCE mode
        circuit_query = (
            sa.select(Circuit)
            .join(Endpoint, Circuit.endpoint_id == Endpoint.id)
            .where(
                sa.and_(
                    Endpoint.health_check_enabled.is_(True), Circuit.app_mode == AppMode.INFERENCE
                )
            )
        )
        circuits = await sess.execute(circuit_query)

        for circuit in circuits.scalars():
            for route in circuit.route_info:
                if route.route_id:  # Only count routes with IDs (health-checkable)
                    total_routes += 1
                    if route.health_status == ModelServiceStatus.HEALTHY:
                        healthy_routes += 1
                    elif route.health_status == ModelServiceStatus.UNHEALTHY:
                        unhealthy_routes += 1
                    else:
                        unknown_routes += 1

        summary = HealthSummaryModel(
            total_endpoints=total_endpoints,
            endpoints_with_health_checks=endpoints_with_health,
            total_routes=total_routes,
            healthy_routes=healthy_routes,
            unhealthy_routes=unhealthy_routes,
            unknown_routes=unknown_routes,
        )

    return PydanticResponse(HealthStatusResponseModel(success=True, summary=summary))


@auth_required("manager")
@pydantic_api_response_handler
async def get_endpoints_health(request: web.Request) -> PydanticResponse[HealthStatusResponseModel]:
    """Get health status for all endpoints"""
    root_ctx: RootContext = request.app["ctx"]

    async with root_ctx.db.begin_readonly_session() as sess:
        # Get all endpoints with health checking enabled
        endpoints = await Endpoint.list_health_check_enabled(sess)

        endpoint_statuses = []
        for endpoint in endpoints:
            # Find the circuit for this endpoint
            try:
                circuit = await Circuit.find_by_endpoint(sess, endpoint.id)

                # Count health statuses from circuit route_info
                healthy_count = 0
                unhealthy_count = 0
                unknown_count = 0
                route_models = []

                for route in circuit.route_info:
                    if route.route_id:  # Only include routes with IDs (health-checkable)
                        if route.health_status == ModelServiceStatus.HEALTHY:
                            healthy_count += 1
                        elif route.health_status == ModelServiceStatus.UNHEALTHY:
                            unhealthy_count += 1
                        else:
                            unknown_count += 1

                        route_models.append(
                            RouteHealthStatusModel(
                                route_id=route.route_id,
                                session_id=route.session_id,
                                kernel_host=route.kernel_host,
                                kernel_port=route.kernel_port,
                                protocol=route.protocol.value,
                                health_status=route.health_status,
                                last_health_check=route.last_health_check,
                                consecutive_failures=route.consecutive_failures,
                                created_at=circuit.created_at,
                                updated_at=circuit.updated_at,
                            )
                        )

                endpoint_status = EndpointHealthStatusModel(
                    endpoint_id=endpoint.id,
                    health_check_enabled=endpoint.health_check_enabled,
                    total_routes=len(route_models),
                    healthy_routes=healthy_count,
                    unhealthy_routes=unhealthy_count,
                    unknown_routes=unknown_count,
                    routes=route_models,
                )
                endpoint_statuses.append(endpoint_status)

            except Exception as e:
                log.error("Failed to find circuit for endpoint {}: {}", endpoint.id, e)
                # Create empty status for endpoints without circuits
                endpoint_status = EndpointHealthStatusModel(
                    endpoint_id=endpoint.id,
                    health_check_enabled=endpoint.health_check_enabled,
                    total_routes=0,
                    healthy_routes=0,
                    unhealthy_routes=0,
                    unknown_routes=0,
                    routes=[],
                )
                endpoint_statuses.append(endpoint_status)

    return PydanticResponse(HealthStatusResponseModel(success=True, endpoints=endpoint_statuses))


@auth_required("manager")
@pydantic_api_response_handler
async def get_circuit_health(
    request: web.Request,
) -> PydanticResponse[HealthStatusResponseModel]:
    """Get health status for a specific circuit"""
    circuit_id_str = request.match_info["circuit_id"]
    try:
        circuit_id = UUID(circuit_id_str)
    except ValueError:
        raise web.HTTPBadRequest(reason="Invalid circuit ID format")

    root_ctx: RootContext = request.app["ctx"]

    async with root_ctx.db.begin_readonly_session() as sess:
        try:
            circuit = await Circuit.get(sess, circuit_id)
        except Exception:
            raise web.HTTPNotFound(reason="Circuit not found")

        if not circuit.endpoint_id:
            # Circuit without endpoint health checking
            circuit_status = CircuitHealthStatusModel(
                circuit_id=circuit.id,
                endpoint_id=None,
                total_routes=len(circuit.route_info),
                healthy_routes=len(circuit.route_info),  # Assume all healthy if no health checking
                all_routes=[],
            )
        else:
            # Get route health from circuit's route_info JSON
            healthy_count = 0
            route_models = []

            for route in circuit.route_info:
                if route.route_id:  # Only include routes with IDs (health-checkable)
                    if route.health_status == ModelServiceStatus.HEALTHY:
                        healthy_count += 1

                    route_models.append(
                        RouteHealthStatusModel(
                            route_id=route.route_id,
                            session_id=route.session_id,
                            kernel_host=route.kernel_host,
                            kernel_port=route.kernel_port,
                            protocol=route.protocol.value,
                            health_status=route.health_status,
                            last_health_check=route.last_health_check,
                            consecutive_failures=route.consecutive_failures,
                            created_at=circuit.created_at,
                            updated_at=circuit.updated_at,
                        )
                    )

            circuit_status = CircuitHealthStatusModel(
                circuit_id=circuit.id,
                endpoint_id=circuit.endpoint_id,
                total_routes=len(route_models),
                healthy_routes=healthy_count,
                all_routes=route_models,
            )

    return PydanticResponse(HealthStatusResponseModel(success=True, circuits=[circuit_status]))


async def hello(request: web.Request) -> web.Response:
    """Simple health check endpoint"""
    from ai.backend.appproxy.common.types import HealthResponse

    request["do_not_print_access_log"] = True

    response = HealthResponse(
        status="healthy",
        version=__version__,
        component="appproxy-coordinator",
    )
    return web.json_response(response.model_dump())


@auth_required("manager")
@pydantic_api_response_handler
async def status(request: web.Request) -> PydanticResponse[StatusResponseModel]:
    """
    Returns health status of coordinator.
    """
    request["do_not_print_access_log"] = True

    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin_readonly_session() as sess:
        workers = await Worker.list_workers(sess)
    return PydanticResponse(
        StatusResponseModel(
            coordinator_version=__version__,
            appproxy_api_version="v2",
            workers=[
                WorkerInfoModel(
                    authority=w.authority,
                    available_slots=w.available_slots,
                    occupied_slots=w.occupied_slots,
                    ha_setup=w.nodes > 1,
                )
                for w in workers
                if (w.updated_at.timestamp() - datetime.now().timestamp()) <= 30
            ],
        )
    )


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "health"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("GET", hello))
    cors.add(add_route("GET", "/status", status))
    cors.add(add_route("GET", "/summary", get_health_summary))
    cors.add(add_route("GET", "/endpoints", get_endpoints_health))
    cors.add(add_route("GET", "/circuits/{circuit_id}", get_circuit_health))
    return app, []



"""rename misrepresentative column names

Revision ID: 6e38326a056a
Revises: e81f833e28fa
Create Date: 2024-03-09 13:26:31.168685

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "6e38326a056a"
down_revision = "e81f833e28fa"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("workers", "available_circuits", new_column_name="available_slots")
    op.alter_column("workers", "occupied_circuits", new_column_name="occupied_slots")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column("workers", "available_slots", new_column_name="available_circuits")
    op.alter_column("workers", "occupied_slots", new_column_name="occupied_circuits")
    # ### end Alembic commands ###



import logging
from typing import Iterable
from uuid import UUID

import aiohttp_cors
from aiohttp import web

from ai.backend.appproxy.common.types import (
    CORSOptions,
    PydanticResponse,
    WebMiddleware,
)
from ai.backend.appproxy.common.utils import (
    pydantic_api_response_handler,
)
from ai.backend.logging import BraceStyleAdapter

from ..models import Circuit
from ..types import RootContext
from .types import StubResponseModel
from .utils import auth_required

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


@auth_required("worker")
@pydantic_api_response_handler
async def clear_circuit(request: web.Request) -> PydanticResponse[StubResponseModel]:
    """
    Breaks the circuit forcefully.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async with root_ctx.db.begin_session() as sess:
        circuit_id = UUID(request.match_info["circuit_id"])
        circuit = await Circuit.get(sess, circuit_id, load_worker=True)

        await root_ctx.circuit_manager.unload_circuits([circuit])

        circuit.worker_row.occupied_slots -= 1
        await sess.delete(circuit)

        await sess.commit()

        log.info(
            "clear_circuit(c: {}): removed circuit by request", request.match_info["circuit_id"]
        )
        return PydanticResponse(StubResponseModel(success=True))


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "v1/api/circuit"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("DELETE", "/{circuit_id}", clear_circuit))
    return app, []



"""separate use_tls into tls_listen and tls_advertised

Revision ID: 7dbbc087108e
Revises: 66f87e010f90
Create Date: 2025-03-31 18:00:24.074776

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql.expression import bindparam

from ai.backend.appproxy.coordinator.models.base import GUID, convention

# revision identifiers, used by Alembic.
revision = "7dbbc087108e"
down_revision = "66f87e010f90"
branch_labels = None
depends_on = None


metadata = sa.MetaData(naming_convention=convention)


def upgrade():
    workers = sa.Table(
        "workers",
        metadata,
        sa.Column("id", GUID),
        sa.Column("use_tls", sa.BOOLEAN()),
        sa.Column("tls_advertised", sa.BOOLEAN()),
    )

    op.add_column(
        "workers", sa.Column("tls_listen", sa.BOOLEAN(), nullable=False, server_default="false")
    )
    op.add_column(
        "workers", sa.Column("tls_advertised", sa.BOOLEAN(), nullable=False, server_default="false")
    )
    conn = op.get_bind()
    query = sa.select(workers)
    updates = [
        {"wid": row.id, "tls_advertised": row.use_tls} for row in conn.execute(query).fetchall()
    ]
    if updates:
        query = (
            sa.update(workers)
            .values(tls_advertised=bindparam("tls_advertised"))
            .where(workers.c.id == bindparam("wid"))
        )
        conn.execute(query, updates)

    op.drop_column("workers", "use_tls")


def downgrade():
    op.add_column(
        "workers", sa.Column("use_tls", sa.BOOLEAN(), autoincrement=False, nullable=False)
    )

    op.drop_column("workers", "tls_advertised")
    op.drop_column("workers", "tls_listen")



"""add worker.wildcard_traffic_port column

Revision ID: a06b7ba8f4bd
Revises: 6e38326a056a
Create Date: 2024-10-22 10:25:40.370766

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a06b7ba8f4bd"
down_revision = "6e38326a056a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workers", sa.Column("wildcard_traffic_port", sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workers", "wildcard_traffic_port")
    # ### end Alembic commands ###



import asyncio
import itertools
import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import (
    Annotated,
    Any,
    AsyncContextManager,
    Callable,
    Optional,
    Protocol,
    Self,
    Sequence,
    TypeAlias,
)
from uuid import UUID

import aiohttp_cors
import attrs
from prometheus_client import generate_latest
from pydantic import AliasChoices, BaseModel, Field, TypeAdapter

from ai.backend.appproxy.common.etcd import TraefikEtcd, convert_to_etcd_dict
from ai.backend.appproxy.common.events import (
    AppProxyCircuitCreatedEvent,
    AppProxyCircuitRemovedEvent,
    AppProxyCircuitRouteUpdatedEvent,
    AppProxyWorkerCircuitAddedEvent,
)
from ai.backend.appproxy.common.exceptions import ServerMisconfiguredError, ServiceUnavailable
from ai.backend.appproxy.common.types import ProxyProtocol, RouteInfo, SerializableCircuit
from ai.backend.appproxy.coordinator.health_checker import HealthCheckEngine
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_schedule import ValkeyScheduleClient
from ai.backend.common.events.dispatcher import EventDispatcher, EventProducer
from ai.backend.common.lock import AbstractDistributedLock
from ai.backend.common.metrics.metric import (
    APIMetricObserver,
    EventMetricObserver,
    SystemMetricObserver,
)
from ai.backend.common.types import AgentId, RedisConnectionInfo
from ai.backend.logging import BraceStyleAdapter

from .config import ServerConfig
from .defs import LockID
from .models import Circuit
from .models.utils import ExtendedAsyncSAEngine

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class CoordinatorMetricRegistry:
    _instance: Optional[Self] = None

    api: APIMetricObserver
    event: EventMetricObserver
    system: SystemMetricObserver

    def __init__(self) -> None:
        self.api = APIMetricObserver.instance()
        self.event = EventMetricObserver.instance()
        self.system = SystemMetricObserver.instance()

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def to_prometheus(self) -> str:
        self.system.observe()
        return generate_latest().decode("utf-8")


class DistributedLockFactory(Protocol):
    def __call__(self, lock_id: LockID, lifetime_hint: float) -> AbstractDistributedLock: ...


@dataclass
class CircuitManager:
    event_dispatcher: EventDispatcher
    event_producer: EventProducer
    traefik_etcd: TraefikEtcd | None
    local_config: ServerConfig

    async def initialize_circuits(self, circuits: Sequence[Circuit]) -> None:
        if self.local_config.proxy_coordinator.enable_traefik:
            await self.initialize_traefik_circuits(circuits)
        else:
            for circuit in circuits:
                await self.initialize_legacy_circuit(circuit)

    async def initialize_traefik_circuits(self, circuits: Sequence[Circuit]) -> None:
        log.debug("initialize_traefik_circuits(): start")
        if not self.traefik_etcd:
            raise ServerMisconfiguredError("proxy-coordinator.traefik")

        for circuit_chunk in itertools.batched(circuits, 5):
            total_map: defaultdict[str, Any] = defaultdict(
                lambda: defaultdict(lambda: {"services": {}, "routers": {}, "middlewares": {}})
            )
            for circuit in circuit_chunk:
                worker_authority = circuit.worker_row.authority

                routers = circuit.traefik_routers
                middlewares = circuit.get_traefik_middlewares(self.local_config)
                services = circuit.traefik_services

                total_map[f"worker_{worker_authority}"][circuit.protocol.value.lower()][
                    "services"
                ].update(services)
                total_map[f"worker_{worker_authority}"][circuit.protocol.value.lower()][
                    "routers"
                ].update(routers)
                if middlewares:
                    total_map[f"worker_{worker_authority}"][circuit.protocol.value.lower()][
                        "middlewares"
                    ].update(middlewares)

            log.debug("traefik_etcd put_prefix {}", convert_to_etcd_dict(total_map))
            await self.traefik_etcd.put_prefix("", convert_to_etcd_dict(total_map))
            log.debug("initialize_traefik_circuits(): end")

    async def initialize_legacy_circuit(self, circuit: Circuit) -> None:
        worker_ready_evt = asyncio.Event()
        authority = circuit.worker_row.authority

        async def _event_handler(
            context: CircuitManager,
            agent_id: AgentId,
            event: AppProxyWorkerCircuitAddedEvent,
        ) -> None:
            if circuit.id in [c.id for c in event.circuits]:
                worker_ready_evt.set()

        worker_ready_event_handler = self.event_dispatcher.subscribe(
            AppProxyWorkerCircuitAddedEvent,
            self,
            _event_handler,
        )
        evt = AppProxyCircuitCreatedEvent(
            authority,
            [SerializableCircuit(**circuit.dump_model())],
        )
        await self.event_producer.broadcast_event(evt)
        try:
            async with asyncio.timeout(15.0):
                await worker_ready_evt.wait()
        except asyncio.TimeoutError:
            raise ServiceUnavailable(
                "E10001: Proxy worker not responding", extra_data={"worker": authority}
            )

        self.event_dispatcher.unsubscribe(worker_ready_event_handler)

    async def update_circuit_routes(self, circuit: Circuit, old_routes: list[RouteInfo]) -> None:
        if self.local_config.proxy_coordinator.enable_traefik:
            await self.update_traefik_circuit_routes(circuit, old_routes)
        else:
            await self.update_legacy_circuit_routes(circuit, old_routes)

    async def update_traefik_circuit_routes(
        self, circuit: Circuit, old_routes: list[RouteInfo]
    ) -> None:
        log.debug("update_traefik_circuit_routes(): start")
        if not self.traefik_etcd:
            raise ServerMisconfiguredError("proxy-coordinator.traefik")

        worker_authority = circuit.worker_row.authority
        etcd_prefix = f"worker_{worker_authority}/{circuit.protocol.value.lower()}"

        # Use health-aware services configuration
        new_route_keys = {
            f"bai_session_{r.session_id}_{circuit.id}" for r in circuit.healthy_routes
        }
        new_services = circuit.traefik_services
        new_services = {
            f"bai_service_{circuit.id}": new_services[f"bai_service_{circuit.id}"],
            **{k: v for k, v in new_services.items() if k in new_route_keys},
        }

        # clear old routes
        for route in old_routes:
            log.debug(
                "traefik_etcd.delete_prefix {}",
                f"{etcd_prefix}/services/bai_session_{route.session_id}_{circuit.id}",
            )
            await self.traefik_etcd.delete_prefix(
                f"{etcd_prefix}/services/bai_session_{route.session_id}_{circuit.id}"
            )
        log.debug(
            "traefik_etcd.delete_prefix {}", f"{etcd_prefix}/services/bai_service_{circuit.id}"
        )
        await self.traefik_etcd.delete_prefix(f"{etcd_prefix}/services/bai_service_{circuit.id}")

        log.debug(
            "traefik_etcd.put_prefix {} {}",
            f"{etcd_prefix}/services",
            convert_to_etcd_dict(new_services),
        )
        await self.traefik_etcd.put_prefix(
            f"{etcd_prefix}/services",
            convert_to_etcd_dict(new_services),
        )
        log.debug("update_traefik_circuit_routes(): end")

    async def update_legacy_circuit_routes(
        self, circuit: Circuit, old_routes: list[RouteInfo]
    ) -> None:
        # Get healthy routes for the circuit
        healthy_routes = circuit.healthy_routes

        event = AppProxyCircuitRouteUpdatedEvent(
            target_worker_authority=circuit.worker_row.authority,
            circuit=SerializableCircuit(**circuit.dump_model()),
            routes=healthy_routes,
        )
        await self.event_producer.broadcast_event(event)

    async def unload_circuits(self, circuits: Sequence[Circuit]) -> None:
        if self.local_config.proxy_coordinator.enable_traefik:
            for circuit in circuits:
                await self.unload_traefik_circuit(circuit)
        else:
            await self.unload_legacy_circuits(circuits)

    async def unload_traefik_circuit(self, circuit: Circuit) -> None:
        log.debug("unload_traefik_circuit(): start")
        if not self.traefik_etcd:
            raise ServerMisconfiguredError("proxy-coordinator.traefik")

        worker_authority = circuit.worker_row.authority
        etcd_prefix = f"worker_{worker_authority}/{circuit.protocol.value.lower()}"

        prefixes_to_remove = [
            f"{etcd_prefix}/routers/bai_router_{circuit.id}",
            f"{etcd_prefix}/services/bai_service_{circuit.id}",
        ] + [
            f"{etcd_prefix}/services/bai_session_{r.session_id}_{circuit.id}"
            for r in circuit.route_info
        ]
        if circuit.protocol == ProxyProtocol.HTTP:
            prefixes_to_remove.append(
                f"{etcd_prefix}/middlewares/bai_appproxy_plugin_{circuit.id}",
            )
            prefixes_to_remove.append(
                f"{etcd_prefix}/middlewares/appproxy/plugin/bai_appproxy_plugin_{circuit.id}",
            )

        for prefix in prefixes_to_remove:
            log.debug("traefik_etcd.delete_prefix {}", prefix)
            await self.traefik_etcd.delete_prefix(prefix)
        log.debug("unload_traefik_circuit(): end")

    async def unload_legacy_circuits(self, circuits: Sequence[Circuit]) -> None:
        circuits_by_worker: defaultdict[str, list[Circuit]] = defaultdict(lambda: [])
        for circuit in circuits:
            circuits_by_worker[circuit.worker_row.authority].append(circuit)

        for authority, circuits in circuits_by_worker.items():
            event = AppProxyCircuitRemovedEvent(
                target_worker_authority=authority,
                circuits=[SerializableCircuit(**c.dump_model()) for c in circuits],
            )
            await self.event_producer.broadcast_event(event)


@attrs.define(slots=True, auto_attribs=True, init=False)
class RootContext:
    pidx: int
    db: ExtendedAsyncSAEngine
    distributed_lock_factory: DistributedLockFactory
    event_dispatcher: EventDispatcher
    event_producer: EventProducer
    core_event_dispatcher: EventDispatcher
    core_event_producer: EventProducer

    valkey_live: ValkeyLiveClient
    redis_lock: RedisConnectionInfo
    core_valkey_live: ValkeyLiveClient
    valkey_schedule: ValkeyScheduleClient
    local_config: ServerConfig
    cors_options: dict[str, aiohttp_cors.ResourceOptions]
    traefik_etcd: TraefikEtcd | None

    circuit_manager: CircuitManager
    health_engine: HealthCheckEngine

    metrics: CoordinatorMetricRegistry


CleanupContext: TypeAlias = Callable[["RootContext"], AsyncContextManager[None]]


class InferenceAppConfig(BaseModel):
    session_id: Annotated[
        UUID,
        Field(
            ...,
            description="ID of the session associated with the inference app.",
            validation_alias=AliasChoices("session-id", "session_id"),
            serialization_alias="session-id",
        ),
    ]
    route_id: Annotated[
        Optional[UUID],
        Field(
            default=None,
            description="ID of the route. This is optional and may not be present for older routes.",
            validation_alias=AliasChoices("route-id", "route_id"),
            serialization_alias="route-id",
        ),
    ]
    kernel_host: Annotated[
        Optional[str],
        Field(
            ...,
            description="Host/IP address of the kernel. This is the address that the proxy will use to connect to the kernel.",
            validation_alias=AliasChoices("kernel-host", "kernel_host"),
            serialization_alias="kernel-host",
        ),
    ]
    kernel_port: Annotated[
        int,
        Field(
            ...,
            ge=1,
            le=65535,
            description="Port number of the kernel. This is the port that the proxy will use to connect to the kernel.",
            validation_alias=AliasChoices("kernel-port", "kernel_port"),
            serialization_alias="kernel-port",
        ),
    ]
    protocol: Annotated[
        ProxyProtocol,
        Field(
            default=ProxyProtocol.HTTP,
            description="Protocol used to connect to the kernel. Supported protocols are HTTP and WebSocket.",
            validation_alias=AliasChoices("protocol"),
            serialization_alias="protocol",
        ),
    ]
    traffic_ratio: Annotated[
        float,
        Field(
            ge=0.0,
            le=1.0,
            default=1.0,
            description="Traffic ratio for the inference app. This is used for load balancing between multiple apps.",
            validation_alias=AliasChoices("traffic-ratio", "traffic_ratio"),
            serialization_alias="traffic-ratio",
        ),
    ]


InferenceAppConfigDict = TypeAdapter(dict[str, list[InferenceAppConfig]])



from datetime import datetime
from typing import TYPE_CHECKING, Any, Optional, Sequence
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import relationship, selectinload
from yarl import URL

from ai.backend.appproxy.common.defs import PERMIT_COOKIE_NAME
from ai.backend.appproxy.common.exceptions import ObjectNotFound, UnsupportedProtocol
from ai.backend.appproxy.common.types import (
    AppMode,
    FrontendMode,
    ProxyProtocol,
    RouteInfo,
    SerializableCircuit,
)
from ai.backend.appproxy.coordinator.config import ServerConfig
from ai.backend.common.types import ModelServiceStatus, RuntimeVariant

from .base import (
    GUID,
    Base,
    BaseMixin,
    EnumType,
    ForeignKeyIDColumn,
    IDColumn,
    StrEnumType,
    StructuredJSONObjectListColumn,
)

if TYPE_CHECKING:
    pass


__all__ = [
    "Circuit",
]


class Circuit(Base, BaseMixin):
    __tablename__ = "circuits"

    """
    Keep information which will be delivered to proxy-worker when establishing
    port-based TCP and/or websocket proxy server.
    """

    id = IDColumn()

    app = sa.Column(sa.String(length=255), nullable=True)
    protocol = sa.Column(EnumType(ProxyProtocol), nullable=False)

    worker = ForeignKeyIDColumn("worker", "workers.id", nullable=False)

    app_mode = sa.Column(EnumType(AppMode), nullable=False)

    frontend_mode = sa.Column(EnumType(FrontendMode), nullable=False)
    port = sa.Column(sa.Integer(), nullable=True)
    subdomain = sa.Column(sa.String(length=255), nullable=True)

    envs = sa.Column(pgsql.JSONB(), nullable=True)
    arguments = sa.Column(sa.String(length=1000), nullable=True)

    open_to_public = sa.Column(sa.Boolean(), nullable=False, default=False)
    allowed_client_ips = sa.Column(sa.String(length=255), nullable=True)

    user_id = sa.Column(GUID, nullable=True)  # null if `app_mode` is set to `INFERENCE`
    endpoint_id = sa.Column(GUID, nullable=True)
    # null if `app_mode` is set to `INTERACTIVE`
    runtime_variant = sa.Column(
        StrEnumType(RuntimeVariant), nullable=True
    )  # null if `app_mode` is set to `INTERACTIVE`

    session_ids = sa.Column(
        pgsql.ARRAY(GUID),
        nullable=False,
        default=[],
    )
    route_info = sa.Column(StructuredJSONObjectListColumn(RouteInfo), nullable=False, default=[])

    created_at = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now())
    updated_at = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now())

    worker_row = relationship("Worker", back_populates="circuits")
    endpoint_row = relationship(
        "Endpoint",
        back_populates="circuit_row",
        primaryjoin="Circuit.endpoint_id == Endpoint.id",
        foreign_keys="Circuit.endpoint_id",
        uselist=False,
    )

    # TODO: Create primary key - worker, port, subdomain

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        circuit_id: UUID,
        load_worker=True,
        load_endpoint=True,
    ) -> "Circuit":
        query = sa.select(Circuit).where((Circuit.id == circuit_id))
        if load_worker:
            query = query.options(selectinload(Circuit.worker_row))
        if load_endpoint:
            query = query.options(selectinload(Circuit.endpoint_row))
        circuit = await session.scalar(query)
        if not circuit:
            raise ObjectNotFound(object_name="circuit")
        return circuit

    @classmethod
    async def get_by_endpoint(
        cls,
        session: AsyncSession,
        endpoint_id: UUID,
        load_worker=True,
        load_endpoint=True,
    ) -> "Circuit":
        query = sa.select(Circuit).where((Circuit.endpoint_id == endpoint_id))
        if load_worker:
            query = query.options(selectinload(Circuit.worker_row))
        if load_endpoint:
            query = query.options(selectinload(Circuit.endpoint_row))
        circuit = await session.scalar(query)
        if not circuit:
            raise ObjectNotFound(object_name="circuit")
        return circuit

    @classmethod
    async def list_circuits(
        cls,
        session: AsyncSession,
        load_worker=True,
        load_endpoint=True,
    ) -> Sequence["Circuit"]:
        query = sa.select(Circuit)
        if load_worker:
            query = query.options(selectinload(Circuit.worker_row))
        if load_endpoint:
            query = query.options(selectinload(Circuit.endpoint_row))
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def find(
        cls,
        session: AsyncSession,
        session_id: UUID,
        app: str,
        open_to_public: bool,
        allowed_client_ips: str | None,
    ) -> "Circuit":
        query = (
            sa.select(Circuit)
            .where(
                (Circuit.session_ids.contains([session_id]))
                & (Circuit.app == app)
                & (Circuit.open_to_public == open_to_public)
                & (Circuit.allowed_client_ips == allowed_client_ips)
            )
            .options(selectinload(Circuit.worker_row))
        )
        circuit = await session.scalar(query)
        if not circuit:
            raise ObjectNotFound(object_name="circuit")
        return circuit

    @classmethod
    async def find_by_endpoint(
        cls,
        session: AsyncSession,
        endpoint_id: UUID,
        load_worker=True,
        load_endpoint=True,
    ) -> "Circuit":
        query = sa.select(Circuit).where((Circuit.endpoint_id == endpoint_id))
        if load_worker:
            query = query.options(selectinload(Circuit.worker_row))
        if load_endpoint:
            query = query.options(selectinload(Circuit.endpoint_row))
        circuit = await session.scalar(query)
        if not circuit:
            raise ObjectNotFound(object_name="circuit")
        return circuit

    @classmethod
    def create(
        cls,
        id: UUID,
        app: str,
        protocol: ProxyProtocol,
        worker: UUID,
        app_mode: AppMode,
        frontend_mode: FrontendMode,
        route_info: list[RouteInfo],
        *,
        port: int | None = None,
        subdomain: str | None = None,
        envs: dict[str, Any] | None = None,
        args: str | None = None,
        open_to_public: bool = False,
        allowed_client_ips: str | None = None,
        user_uuid: UUID | None = None,
        endpoint_id: UUID | None = None,
        runtime_variant: RuntimeVariant | None = None,
    ) -> "Circuit":
        c = Circuit()
        c.id = id
        c.app = app
        c.protocol = protocol
        c.worker = worker
        c.app_mode = app_mode
        c.frontend_mode = frontend_mode
        c.port = port
        c.subdomain = subdomain
        c.envs = envs
        c.arguments = args
        c.open_to_public = open_to_public
        c.allowed_client_ips = allowed_client_ips
        c.user_id = user_uuid
        c.endpoint_id = endpoint_id
        c.route_info = route_info
        c.session_ids = [r.session_id for r in route_info]
        c.runtime_variant = runtime_variant

        c.created_at = datetime.now()
        c.updated_at = datetime.now()

        return c

    async def get_endpoint_url(self, session: Optional[AsyncSession] = None) -> URL:
        from .worker import Worker

        worker: Worker = (
            await Worker.get(session, self.worker) if session is not None else self.worker_row
        )

        match (worker.use_tls, self.protocol):
            case (True, ProxyProtocol.TCP):
                scheme = "tls"
            case (False, ProxyProtocol.TCP):
                scheme = "tcp"
            case (_, ProxyProtocol.GRPC):
                scheme = "grpc"
            case (True, ProxyProtocol.HTTP):
                scheme = "https"
            case (False, ProxyProtocol.HTTP):
                scheme = "http"
            case _:
                raise UnsupportedProtocol(self.protocol.name)

        match self.frontend_mode:
            case FrontendMode.WILDCARD_DOMAIN:
                assert self.subdomain and worker.wildcard_domain
                hostname = self.subdomain + worker.wildcard_domain
                if (scheme == "https" and worker.wildcard_traffic_port != 443) or (
                    scheme == "http" and worker.wildcard_traffic_port != 80
                ):
                    hostname += f":{worker.wildcard_traffic_port}"
            case FrontendMode.PORT:
                assert self.port
                hostname = f"{worker.hostname}:{self.port}"

        url = f"{scheme}://{hostname}"
        return URL(url)

    async def get_slot_identity(self) -> str:
        from .worker import Worker

        worker: Worker = self.worker_row
        return f"{worker.authority}:{self.port or self.subdomain}"

    @property
    def traefik_rule(self) -> str:
        match self.protocol:
            case ProxyProtocol.TCP:
                return "ClientIP(`0.0.0.0/0`)"
            case _:
                match self.frontend_mode:
                    case FrontendMode.PORT:
                        return f"Host(`{self.worker_row.hostname}`)"
                    case FrontendMode.WILDCARD_DOMAIN:
                        assert self.subdomain
                        return f"Host(`{self.subdomain}{self.worker_row.wildcard_domain}`)"
                    case _:
                        raise ValueError(
                            f"Invalid frontend mode for traefik setup: {self.frontend_mode}"
                        )

    @property
    def traefik_entrypoint(self) -> str:
        match self.frontend_mode:
            case FrontendMode.WILDCARD_DOMAIN:
                return "domainproxy"
            case FrontendMode.PORT:
                assert self.port
                return f"portproxy_{self.port}"
            case _:
                raise ValueError(f"Invalid frontend mode for traefik setup: {self.frontend_mode}")

    @property
    def traefik_routers(self) -> dict[str, Any]:
        match self.protocol:
            case ProxyProtocol.HTTP:
                base = {
                    "rule": self.traefik_rule,
                    "service": f"bai_service_{self.id}",
                    "entrypoints": [self.traefik_entrypoint],
                    "middlewares": [
                        "CORSHeaders",
                        f"bai_appproxy_plugin_{self.id}",
                        f"bai_appproxy_plugin_{self.id}_go",
                    ],
                }
                if self.worker_row.tls_listen:
                    base["tls"] = ""
                return {f"bai_router_{self.id}": base}
            case ProxyProtocol.TCP:
                return {
                    f"bai_router_{self.id}": {
                        "rule": self.traefik_rule,
                        "service": f"bai_service_{self.id}",
                        "entrypoints": [self.traefik_entrypoint],
                    },
                }
        return {}

    @property
    def healthy_routes(self) -> list[RouteInfo]:
        """
        Get only healthy routes for this circuit.

        Health filtering is only applied for circuits in INFERENCE mode.
        For other modes, all routes are considered healthy.

        Returns:
            List of healthy RouteInfo objects
        """
        # Only apply health filtering for circuits in INFERENCE mode
        if self.app_mode != AppMode.INFERENCE or not self.endpoint_id:
            # No health filtering, return all routes
            return self.route_info

        assert self.endpoint_row
        # Filter routes based on health status stored in JSON
        healthy_routes = []
        for route in self.route_info:
            # Include routes that are explicitly healthy or have no health status (not health-checked)
            if (
                not self.endpoint_row.health_check_enabled
                or route.health_status == ModelServiceStatus.HEALTHY
            ):
                healthy_routes.append(route)

        return healthy_routes

    def update_route_health_status(
        self,
        route_id: UUID,
        health_status: ModelServiceStatus | None,
        last_check_time: float | None = None,
        consecutive_failures: int | None = None,
    ) -> bool:
        """
        Update health status for a specific route in the circuit's route_info

        Args:
            route_id: ID of the route to update
            health_status: New health status
            last_check_time: Timestamp of last health check
            consecutive_failures: Number of consecutive failures

        Returns:
            True if route was found and updated, False otherwise
        """
        for route in self.route_info:
            if route.route_id == route_id:
                did_update_status = False
                if route.health_status != health_status:
                    route.health_status = health_status
                    did_update_status = True
                if last_check_time is not None:
                    route.last_health_check = last_check_time
                if consecutive_failures is not None:
                    route.consecutive_failures = consecutive_failures

                # Mark the route_info column as modified for SQLAlchemy change tracking
                # This is necessary because SQLAlchemy doesn't automatically detect
                # changes to nested objects within JSON columns
                from sqlalchemy.orm import attributes

                attributes.flag_modified(self, "route_info")

                return did_update_status
        return False

    @property
    def traefik_services(self) -> dict[str, Any]:
        # Use health-aware route filtering
        healthy_routes = self.healthy_routes

        # If no healthy routes, return empty config to remove from load balancer
        if not healthy_routes:
            return {}

        base = {  # services should be inserted separately to prevent overwritting whole `services` prefix
            f"bai_service_{self.id}": {
                "weighted": {
                    "services": [
                        {
                            "name": f"bai_session_{r.session_id}_{self.id}",
                            "weight": int(r.traffic_ratio),
                        }
                        for r in healthy_routes
                    ]
                }
            },
        }
        match self.protocol:
            case ProxyProtocol.HTTP:
                for r in healthy_routes:
                    base.update({
                        f"bai_session_{r.session_id}_{self.id}": {
                            "loadBalancer": {
                                "servers": [
                                    {"url": f"http://{r.current_kernel_host}:{r.kernel_port}/"}
                                ],
                            }
                        }
                    })
            case ProxyProtocol.TCP:
                for r in healthy_routes:
                    base.update({
                        f"bai_session_{r.session_id}_{self.id}": {
                            "loadBalancer": {
                                "servers": [
                                    {"address": f"{r.current_kernel_host}:{r.kernel_port}"}
                                ],
                            }
                        }
                    })
        return base

    def get_traefik_middlewares(self, local_config: ServerConfig) -> dict[str, Any]:
        match self.protocol:
            case ProxyProtocol.HTTP:
                traefik_config = local_config.proxy_coordinator.traefik
                assert traefik_config

                return {
                    "CORSHeaders": {
                        "headers": {
                            "accessControlAllowHeaders": "*",
                            "accessControlAllowOriginList": ["*"],
                        },
                    },
                    f"bai_appproxy_plugin_{self.id}": {
                        "plugin": {
                            "appproxy-traefik-plugin": {
                                # Etcd cannot represent bool or None so we're going to just dump the whole JSON
                                "circuit": SerializableCircuit(
                                    **self.dump_model()
                                ).model_dump_json(),
                                "jwt_secret": local_config.secrets.jwt_secret,
                                "permit_hash_secret": local_config.permit_hash.secret.decode(),
                                "permit_cookie_name": PERMIT_COOKIE_NAME,
                            },
                        }
                    },
                    f"bai_appproxy_plugin_{self.id}_go": {
                        "plugin": {
                            "appproxy-traefik-plugin-go": {
                                "id": str(self.id),
                                "sessionids": [str(x) for x in self.session_ids],
                                "lastusedmarkerpath": self.worker_row.traefik_last_used_marker_path,
                            }
                        }
                    },
                }
        return {}



import logging
import uuid
from enum import StrEnum
from typing import TYPE_CHECKING, Any
from uuid import UUID

if TYPE_CHECKING:
    pass

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pgsql
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import relationship, selectinload
from yarl import URL

from ai.backend.appproxy.common.exceptions import (
    ObjectNotFound,
    PortNotAvailable,
    WorkerNotAvailable,
)
from ai.backend.appproxy.common.types import (
    AppMode,
    EndpointConfig,
    FrontendMode,
    ProxyProtocol,
    RouteInfo,
    SessionConfig,
    Slot,
)
from ai.backend.logging import BraceStyleAdapter

from .base import Base, BaseMixin, EnumType, ForeignKeyIDColumn, IDColumn, StrEnumType
from .circuit import Circuit

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

__all__ = [
    "Worker",
    "WorkerAppFilter",
    "WorkerStatus",
    "pick_worker",
    "add_circuit",
]


class WorkerStatus(StrEnum):
    ALIVE = "ALIVE"
    LOST = "LOST"
    TERMINATED = "TERMINATED"


class Worker(Base, BaseMixin):
    __tablename__ = "workers"

    id = IDColumn()

    authority = sa.Column(
        sa.String(length=255), nullable=False, unique=True
    )  # Human-readable identity of each AppProxy worker, must be identical across all AppProxy workers under single VIP when HA is set up
    frontend_mode = sa.Column(
        EnumType(FrontendMode), nullable=False
    )  # will be fixed as `port` when `protocol` is set to `tcp`
    protocol = sa.Column(EnumType(ProxyProtocol), nullable=False)  # type of traffic to procy

    hostname = sa.Column(
        sa.String(length=1024), nullable=False
    )  # Hostname which users utilize to access this AppProxy
    tls_listen = sa.Column(
        sa.Boolean(), default=False
    )  # Indicates if TLS is required to access the AppProxy
    tls_advertised = sa.Column(
        sa.Boolean(), default=False
    )  # Indicates if TLS is required to access the AppProxy

    api_port = sa.Column(sa.Integer(), nullable=False)  # REST API port

    available_slots = sa.Column(
        sa.Integer(), default=0, nullable=False
    )  # set to -1 when `frontend_mode` is set to `wildcard`
    occupied_slots = sa.Column(sa.Integer(), default=0, nullable=False)

    # Only set if `frontend_mode` is `port`
    port_range = sa.Column(pgsql.ARRAY(sa.Integer), nullable=True)

    # Only set if `frontend_mode` is `wildcard`
    # .example.com
    wildcard_domain = sa.Column(sa.String(length=1024), nullable=True)
    # Only set if `frontend_mode` is `wildcard`
    wildcard_traffic_port = sa.Column(sa.Integer(), nullable=True)

    nodes = sa.Column(sa.Integer(), default=1, nullable=False)

    accepted_traffics = sa.Column(pgsql.ARRAY(EnumType(AppMode)), nullable=False)
    filtered_apps_only = sa.Column(sa.Boolean(), default=False, nullable=False)

    traefik_last_used_marker_path = sa.Column(sa.String(length=1024), nullable=True)

    created_at = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now())
    updated_at = sa.Column(sa.DateTime(timezone=True), server_default=sa.func.now())

    status = sa.Column(
        StrEnumType(WorkerStatus),
        default=WorkerStatus.ALIVE,
        nullable=False,
    )

    filters = relationship(
        "WorkerAppFilter",
        back_populates="worker_row",
    )
    circuits = relationship(
        "Circuit",
        back_populates="worker_row",
    )

    @classmethod
    async def get(
        cls,
        session: AsyncSession,
        worker_id: UUID,
        load_filters=False,
        load_circuits=False,
    ) -> "Worker":
        query = sa.select(Worker).filter(Worker.id == worker_id)
        if load_filters:
            query = query.options(selectinload(Worker.filters))
        if load_circuits:
            query = query.options(selectinload(Worker.circuits))
        worker = await session.scalar(query)
        if not worker:
            raise ObjectNotFound(object_name="worker")
        return worker

    @classmethod
    async def find_by_authority(
        cls,
        session: AsyncSession,
        authority: str,
        load_filters=False,
        load_circuits=False,
    ) -> "Worker":
        query = sa.select(Worker).filter(Worker.authority == authority)
        if load_filters:
            query = query.options(selectinload(Worker.filters))
        if load_circuits:
            query = query.options(selectinload(Worker.circuits))
        worker = await session.scalar(query)
        if not worker:
            raise ObjectNotFound(object_name="worker")
        return worker

    @classmethod
    async def list_workers(
        cls,
        session: AsyncSession,
        load_filters=False,
        load_circuits=False,
    ) -> list["Worker"]:
        query = sa.select(Worker)
        if load_filters:
            query = query.options(selectinload(Worker.filters))
        if load_circuits:
            query = query.options(selectinload(Worker.circuits))
        worker = await session.scalars(query)
        return list(worker)

    @classmethod
    def create(
        cls,
        id: UUID,
        authority: str,
        frontend_mode: FrontendMode,
        protocol: ProxyProtocol,
        hostname: str,
        tls_listen: bool,
        tls_advertised: bool,
        api_port: int,
        accepted_traffics: list[AppMode],
        *,
        port_range: tuple[int, int] | None = None,
        wildcard_domain: str | None = None,
        wildcard_traffic_port: int | None = None,
        traefik_last_used_marker_path: str | None = None,
        filtered_apps_only=False,
        status=WorkerStatus.LOST,
    ) -> "Worker":
        w = cls()
        w.id = id
        w.authority = authority
        w.frontend_mode = frontend_mode
        w.protocol = protocol
        w.hostname = hostname
        w.tls_listen = tls_listen
        w.tls_advertised = tls_advertised
        w.api_port = api_port
        w.accepted_traffics = accepted_traffics
        w.port_range = port_range
        w.wildcard_domain = wildcard_domain
        w.wildcard_traffic_port = wildcard_traffic_port
        w.filtered_apps_only = filtered_apps_only
        w.traefik_last_used_marker_path = traefik_last_used_marker_path
        w.status = status

        w.occupied_slots = 0
        match frontend_mode:
            case FrontendMode.WILDCARD_DOMAIN:
                assert wildcard_domain
                w.available_slots = -1
            case FrontendMode.PORT:
                assert port_range
                w.available_slots = port_range[1] - port_range[0] + 1

        return w

    @property
    def use_tls(self) -> bool:
        return self.tls_listen or self.tls_advertised

    @property
    def api_endpoint(self) -> URL:
        should_include_port = (not self.use_tls and self.api_port != 80) or (
            self.use_tls and self.api_port != 443
        )
        base_url = f"{'https' if self.use_tls else 'http'}://{self.hostname}"
        if should_include_port:
            base_url += f":{self.api_port}"
        return URL(base_url)

    async def list_slots(self, session: AsyncSession) -> list[Slot]:
        """
        For workers working with PORT-based frontend_mode, this will list all available and occupied slots
        For workers with SUBDOMAIN, this will list occupied slots only - we can't list all available slots since it is infinite
        """
        circuit_list_query = sa.select(Circuit).where(Circuit.worker == self.id)
        circuits: list[Circuit] = (await session.execute(circuit_list_query)).scalars().all()
        match self.frontend_mode:
            case FrontendMode.PORT:
                assert self.port_range
                occupied_ports: dict[int, UUID] = {c.port: c.id for c in circuits if c.port}
                slots = [
                    Slot(
                        FrontendMode.PORT,
                        port in occupied_ports,
                        port=port,
                        subdomain=None,
                        circuit_id=occupied_ports.get(port),
                    )
                    for port in range(self.port_range[0], self.port_range[1] + 1)
                ]
            case FrontendMode.WILDCARD_DOMAIN:
                slots = [
                    Slot(
                        FrontendMode.WILDCARD_DOMAIN,
                        True,
                        port=None,
                        subdomain=c.subdomain,
                        circuit_id=c.id,
                    )
                    for c in circuits
                ]
            case _:
                raise ValueError(f"Invalid frontend mode: {self.frontend_mode}")
        return slots


class WorkerAppFilter(Base, BaseMixin):
    __tablename__ = "worker_app_filters"

    id = IDColumn()

    property_name = sa.Column(sa.VARCHAR(96), nullable=False)
    property_value = sa.Column(sa.VARCHAR(1024), nullable=False)
    worker = ForeignKeyIDColumn("worker", "workers.id", nullable=False)

    worker_row = relationship("Worker", back_populates="filters")

    __table_args__ = (
        sa.UniqueConstraint(
            "worker",
            "property_name",
            "property_value",
            name="uq_worker_app_filter_worker_property_name_property_value",
        ),
    )

    @classmethod
    def create(
        cls, id: UUID, property_name: str, property_value: str, worker: UUID
    ) -> "WorkerAppFilter":
        filter = WorkerAppFilter()
        filter.id = id
        filter.property_name = property_name
        filter.property_value = property_value
        filter.worker = worker
        return filter

    @classmethod
    async def find_by_rule(
        cls, session: AsyncSession, worker: UUID, name: str, value: str
    ) -> "WorkerAppFilter":
        query = sa.select(WorkerAppFilter).where(
            (WorkerAppFilter.worker == worker)
            & (WorkerAppFilter.property_name == name)
            & (WorkerAppFilter.property_value == value)
        )
        rule = await session.scalar(query)
        if not rule:
            raise ObjectNotFound(object_name="worker_app_filter")
        return rule


async def pick_worker(
    session: AsyncSession,
    session_info: SessionConfig,
    endpoint_info: EndpointConfig | None,
    protocol: ProxyProtocol,
    app_mode: AppMode,
) -> Worker:
    app_filter_queries = [
        (WorkerAppFilter.property_name == f"session.{key}")
        & (WorkerAppFilter.property_value == str(value))
        for key, value in session_info.model_dump().items()
    ]
    if endpoint_info:
        app_filter_queries += [
            (WorkerAppFilter.property_name == f"endpoint.{key}")
            & (WorkerAppFilter.property_value == str(value))
            for key, value in endpoint_info.model_dump().items()
        ]
    where_clause = app_filter_queries[0]
    for q in app_filter_queries[1:]:
        where_clause |= q
    query = sa.select(WorkerAppFilter).where(where_clause)
    filter_result = await session.execute(query)
    worker_ids = [app_filter.worker for app_filter in filter_result.scalars().all()]

    log.debug("protocol: {} ({})", protocol, type(protocol))
    worker_query = sa.select(Worker).where(
        (Worker.protocol == protocol)
        & (Worker.accepted_traffics.contains([app_mode]))
        & (Worker.status == WorkerStatus.ALIVE)
    )
    if worker_ids:
        worker_query = worker_query.where(
            Worker.filtered_apps_only & (Worker.accepted_traffics.contains(app_mode))
        )
    result = await session.execute(worker_query)
    sorted_workers: list[Worker] = [
        f
        for f in sorted(
            result.scalars().all(),
            key=lambda f: (
                -1
                if f.frontend_mode == FrontendMode.WILDCARD_DOMAIN
                else (f.available_slots - f.occupied_slots) * -1
            ),
            reverse=True,
        )
        if f.frontend_mode == FrontendMode.WILDCARD_DOMAIN
        or (f.available_slots - f.occupied_slots) > 0
    ]
    if not sorted_workers:
        raise WorkerNotAvailable
    worker = sorted_workers[0]
    return await Worker.get(
        session,
        worker.id,
        load_circuits=True,
    )


async def add_circuit(
    session: AsyncSession,
    session_info: SessionConfig,
    endpoint_info: EndpointConfig | None,
    app: str,
    protocol: ProxyProtocol,
    mode: AppMode,
    routes: list[RouteInfo],
    *,
    envs: dict[str, Any] = {},
    args: str | None = None,
    open_to_public=False,
    allowed_client_ips: str | None = None,
    preferred_port: int | None = None,
    preferred_subdomain: str | None = None,
    worker_id: UUID | None = None,
) -> tuple[Circuit, Worker]:
    if worker_id:
        worker = await Worker.get(session, worker_id, load_circuits=True)
        if worker.available_slots - worker.occupied_slots <= 0 and worker.available_slots >= 0:
            raise WorkerNotAvailable
    else:
        worker = await pick_worker(session, session_info, endpoint_info, protocol, mode)

    circuit_params: dict[str, Any] = {}

    if worker.frontend_mode == FrontendMode.WILDCARD_DOMAIN:
        acquired_subdomains = [c.subdomain for c in worker.circuits]
        if _requested_subdomain := preferred_subdomain:
            subdomain = _requested_subdomain
        else:
            sub_id = str(uuid.uuid4()).split("-")[0]
            subdomain = f"app-{sub_id}"

        while subdomain in acquired_subdomains:
            sub_id = str(uuid.uuid4()).split("-")[0]
            subdomain = f"{_requested_subdomain}-{sub_id}"
        circuit_params["subdomain"] = subdomain
    else:
        acquired_ports = set([c.port for c in worker.circuits])
        port_range = worker.port_range
        assert port_range
        port_pool = set(range(port_range[0], port_range[1] + 1)) - acquired_ports
        if _requested_port := preferred_port:
            if _requested_port not in port_pool:
                raise PortNotAvailable
            port = _requested_port
        else:
            if len(port_pool) == 0:
                raise PortNotAvailable
            port = port_pool.pop()
        circuit_params["port"] = port

    circuit = Circuit.create(
        uuid.uuid4(),
        app,
        protocol,
        worker.id,
        mode,
        worker.frontend_mode,
        routes,
        envs=envs,
        args=args,
        open_to_public=open_to_public,
        allowed_client_ips=allowed_client_ips,
        user_uuid=session_info.user_uuid,
        endpoint_id=endpoint_info.id if endpoint_info else None,
        runtime_variant=endpoint_info.runtime_variant if endpoint_info else None,
        **circuit_params,
    )
    circuit.worker_row = worker

    worker.occupied_slots += 1
    session.add(circuit)
    return (circuit, worker)



from __future__ import annotations

import logging
import urllib.parse
from typing import TYPE_CHECKING, Annotated, Iterable, Optional
from uuid import UUID

import jwt
from aiohttp import web
from pydantic import AnyUrl, BaseModel, Field

from ai.backend.appproxy.common.exceptions import (
    InvalidCredentials,
    ObjectNotFound,
)
from ai.backend.appproxy.common.types import (
    AppMode,
    CORSOptions,
    ProxyProtocol,
    PydanticResponse,
    RouteInfo,
    SessionConfig,
    WebMiddleware,
)
from ai.backend.appproxy.common.utils import mime_match, pydantic_api_handler
from ai.backend.appproxy.coordinator.api.types import ConfRequestModel
from ai.backend.logging import BraceStyleAdapter

from ..models import Circuit, Token, Worker, add_circuit
from ..models.utils import execute_with_txn_retry
from ..types import RootContext

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession as SASession

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class AddRequestModel(BaseModel):
    app: str
    protocol: ProxyProtocol
    envs: Annotated[dict[str, str | int | None], Field(default={})]
    args: Annotated[str | None, Field(default=None)]
    open_to_public: Annotated[bool, Field(default=False)]
    allowed_client_ips: Annotated[str | None, Field(default=None)]
    redirect: Annotated[str, Field(default="")]
    no_reuse: Annotated[bool, Field(default=False)]

    port: Annotated[int | None, Field(default=None)]
    subdomain: Annotated[str | None, Field(default=None)]


class ProxyRequestModel(AddRequestModel):
    token: str
    session_id: UUID


class AddResponseModel(BaseModel):
    code: int
    url: AnyUrl


class ProxyResponseModel(BaseModel):
    redirect_url: AnyUrl
    reuse: bool


class TokenBodyModel(ConfRequestModel):
    exp: int


@pydantic_api_handler(AddRequestModel, is_deprecated=True)
async def add(request: web.Request, params: AddRequestModel) -> PydanticResponse[AddResponseModel]:
    """
    Deprecated: only for legacy applications. Just call `proxy` API directly.
    Returns URL to AppProxy's `proxy` API handler.
    """
    root_ctx: RootContext = request.app["_root.context"]

    coordinator_config = root_ctx.local_config.proxy_coordinator
    base_url = coordinator_config.advertise_base_url
    qdict = {
        **params.model_dump(mode="json", exclude_defaults=True),
        "token": request.match_info["token"],
        "session_id": request.match_info["session_id"],
    }
    return PydanticResponse(
        AddResponseModel(
            code=200, url=AnyUrl(f"{base_url}/v2/proxy/auth?{urllib.parse.urlencode(qdict)}")
        ),
        headers={"Access-Control-Allow-Origin": "*", "Access-Control-Expose-Headers": "*"},
    )


@pydantic_api_handler(ProxyRequestModel)
async def proxy(
    request: web.Request, params: ProxyRequestModel
) -> PydanticResponse[ProxyResponseModel] | web.HTTPPermanentRedirect:
    """
    Assigns worker to host proxy app and starts proxy process.
    When `Accept` HTTP header is set to `application/json` access information to worker will be handed out inside response body;
    otherwise coordinator will try to automatically redirect callee via `Location: ` response header.
    """

    existing_circuit: Optional[Circuit] = None
    reuse = False

    root_ctx: RootContext = request.app["_root.context"]
    token_id = params.token
    session_id = params.session_id

    async with root_ctx.db.begin_readonly_session() as sess:
        token = await Token.get(sess, UUID(token_id))

    if token.session_id != session_id:
        log.warning(
            "User requested to create app of session {} but token authorizes session {}",
            session_id,
            token.session_id,
        )
        raise InvalidCredentials("E20007: Session ID mismatch")

    if not params.no_reuse:
        async with root_ctx.db.begin_readonly_session() as sess:
            try:
                existing_circuit = await Circuit.find(
                    sess,
                    token.session_id,
                    params.app,
                    params.open_to_public,
                    params.allowed_client_ips,
                )
            except ObjectNotFound:
                existing_circuit = None

    if existing_circuit:
        reuse = True
        routes = existing_circuit.route_info
        async with root_ctx.db.begin_readonly_session() as sess:
            worker = await Worker.get(sess, existing_circuit.worker)
        log.debug("reusing existing circuit {}", existing_circuit.id)
        circuit = existing_circuit
    else:
        routes = [
            RouteInfo(
                session_id=token.session_id,
                session_name=None,
                kernel_host=token.kernel_host,
                kernel_port=token.kernel_port,
                protocol=params.protocol,
                traffic_ratio=1.0,
                route_id=None,
                health_status=None,
                last_health_check=None,
                consecutive_failures=0,
            )
        ]

        log.debug("protocol: {} ({})", params.protocol, type(params.protocol))
        if params.protocol == ProxyProtocol.PREOPEN:
            log.debug("overriding PREOPEN to HTTP")
            params.protocol = ProxyProtocol.HTTP

        async def _update(sess: SASession) -> tuple[Circuit, Worker]:
            circuit, worker = await add_circuit(
                sess,
                SessionConfig(
                    id=token.session_id,
                    user_uuid=token.user_uuid,
                    group_id=token.group_id,
                    access_key=token.access_key,
                    domain_name=token.domain_name,
                ),
                None,
                params.app,
                params.protocol,
                AppMode.INTERACTIVE,
                routes,
                envs=params.envs,
                args=params.args,
                open_to_public=params.open_to_public,
                allowed_client_ips=params.allowed_client_ips,
                preferred_port=params.port,
                preferred_subdomain=params.subdomain,
            )
            return circuit, worker

        async with root_ctx.db.connect() as db_conn:
            circuit, worker = await execute_with_txn_retry(
                _update, root_ctx.db.begin_session, db_conn
            )
        log.debug("created new circuit {}", circuit.id)

    assert circuit and worker

    await root_ctx.circuit_manager.initialize_circuits([circuit])

    assert circuit
    log.debug("Circuit is set (id:{})", str(circuit.id))
    token_to_generate_body = {
        "id": str(token.id),
        "version": "v2",  # TODO: add support for v1
        "redirect": params.redirect,
        "circuit": str(circuit.id),
    }
    qdict = {
        "token": jwt.encode(token_to_generate_body, root_ctx.local_config.secrets.jwt_secret),
    }

    app_url = f"{worker.api_endpoint}/setup?{urllib.parse.urlencode(qdict)}"
    log.debug("Redirect URL created: {}", app_url)

    if mime_match(request.headers.get("accept", "text/html"), "application/json", strict=True):
        # Web browsers block redirect between cross-origins if Access-Control-Allow-Origin value is set to a concrete Origin instead of wildcard;
        # Hence we need to send "*" as allowed origin manually, instead of benefiting from aiohttp-cors
        return PydanticResponse(
            ProxyResponseModel(
                redirect_url=AnyUrl(app_url),
                # Current version of WebUI always expects this to be False for TCP protocols
                reuse=reuse if params.protocol != ProxyProtocol.TCP else False,
            ),
            headers={"Access-Control-Allow-Origin": "*", "Access-Control-Expose-Headers": "*"},
        )
    else:
        return web.HTTPPermanentRedirect(app_url)


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "v2/proxy"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app.router.add_route("GET", "/{token}/{session_id}/add", add)
    app.router.add_route("GET", "/auth", proxy)
    return app, []



"""create circuits table

Revision ID: e81f833e28fa
Revises: 5c6942374c62
Create Date: 2024-02-23 17:34:15.169451

"""

import enum

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from ai.backend.appproxy.common.types import RouteInfo
from ai.backend.appproxy.coordinator.models.base import (
    GUID,
    StructuredJSONObjectListColumn,
)

# revision identifiers, used by Alembic.
revision = "e81f833e28fa"
down_revision = "5c6942374c62"
branch_labels = None
depends_on = None


class FrontendMode(enum.StrEnum):
    WILDCARD_DOMAIN = "wildcard"
    PORT = "port"


class ProxyProtocol(enum.StrEnum):
    HTTP = "http"
    GRPC = "grpc"
    HTTP2 = "h2"
    TCP = "tcp"


class AppMode(enum.StrEnum):
    INTERACTIVE = "interactive"
    INFERENCE = "inference"


appmode_choices = [v.name for v in AppMode]
appmode = postgresql.ENUM(*appmode_choices, name="appmode", create_type=False)

frontendmode_choices = [v.name for v in FrontendMode]
frontendmode = postgresql.ENUM(*frontendmode_choices, name="frontendmode", create_type=False)

proxyprotocol_choices = [v.name for v in ProxyProtocol]
proxyprotocol = postgresql.ENUM(*proxyprotocol_choices, name="proxyprotocol", create_type=False)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "circuits",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("app", sa.String(length=255), nullable=True),
        sa.Column("protocol", proxyprotocol, nullable=False),
        sa.Column("worker", GUID(), nullable=False),
        sa.Column("app_mode", appmode, nullable=False),
        sa.Column("frontend_mode", frontendmode, nullable=False),
        sa.Column("port", sa.Integer(), nullable=True),
        sa.Column("subdomain", sa.String(length=255), nullable=True),
        sa.Column("envs", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
        sa.Column("arguments", sa.String(length=1000), nullable=True),
        sa.Column("open_to_public", sa.Boolean(), nullable=False),
        sa.Column("allowed_client_ips", sa.String(length=255), nullable=True),
        sa.Column("user_id", GUID(), nullable=True),
        sa.Column("endpoint_id", GUID(), nullable=True),
        sa.Column("session_ids", sa.ARRAY(GUID()), nullable=False),
        sa.Column("route_info", StructuredJSONObjectListColumn(RouteInfo), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["worker"], ["workers.id"], name=op.f("fk_circuits_worker_workers")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_circuits")),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("circuits")
    # ### end Alembic commands ###



from __future__ import annotations

import asyncio
from logging.config import fileConfig

from alembic import context
from sqlalchemy import pool
from sqlalchemy.engine import Connection
from sqlalchemy.ext.asyncio import async_engine_from_config

from ai.backend.appproxy.coordinator.models.alembic import invoked_programmatically
from ai.backend.appproxy.coordinator.models.base import metadata_obj as metadata
from ai.backend.logging import is_active as logging_active

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.

if not logging_active.get():
    assert config.config_file_name is not None
    fileConfig(config.config_file_name)

# Import the shared metadata and all models.
# (We need to explicilty import models because model modules
# should be executed to add table definitions to the metadata.)

target_metadata = metadata

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(url=url, target_metadata=target_metadata, literal_binds=True)

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: Connection) -> None:
    context.configure(connection=connection, target_metadata=target_metadata)

    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations() -> None:
    connectable = async_engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )
    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)
    await connectable.dispose()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.
    """
    connection = config.attributes.get("connection", None)
    if connection is None:
        asyncio.run(run_async_migrations())
    else:
        do_run_migrations(connection)


if not invoked_programmatically.get():  # when executed via `alembic` commands
    if context.is_offline_mode():
        run_migrations_offline()
    else:
        run_migrations_online()



import asyncio
import importlib
import json
import logging
import subprocess
import sys
from pathlib import Path
from typing import Any, Optional

import aiohttp_cors
import click
import tomli_w
from aiohttp import web
from setproctitle import setproctitle

from ai.backend.appproxy.common.config import generate_example_json
from ai.backend.appproxy.common.openapi import generate_openapi
from ai.backend.appproxy.common.utils import ensure_json_serializable
from ai.backend.cli.types import ExitCode
from ai.backend.logging import BraceStyleAdapter, LogLevel

from ..config import ServerConfig
from .context import CLIContext

log = BraceStyleAdapter(logging.getLogger("ai.backend.manager.cli"))


@click.group(invoke_without_command=False, context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        exists=True,
        path_type=Path,
    ),
    default=None,
    help="The config file path. (default: ./manager.conf and /etc/backend.ai/manager.conf)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Set the logging level to DEBUG",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    debug: bool,
    log_level: LogLevel,
    config_path: Optional[Path] = None,
) -> None:
    """
    Proxy Coordinator Administration CLI
    """
    setproctitle("backend.ai: proxy-coordinator.cli")
    if debug:
        log_level = LogLevel.DEBUG
    ctx.obj = ctx.with_resource(CLIContext(config_path=config_path, log_level=log_level))


@main.command()
@click.option(
    "--output",
    "-o",
    default="-",
    type=click.Path(dir_okay=False, writable=True),
    help="Output file path (default: stdout)",
)
def generate_example_configuration(output: Path) -> None:
    """
    Generates example TOML configuration file for Backend.AI Proxy Coordinator.
    """
    if output == "-" or output is None:
        print(tomli_w.dumps(ensure_json_serializable(generate_example_json(ServerConfig))))
    else:
        with open(output, mode="w") as fw:
            fw.write(tomli_w.dumps(ensure_json_serializable(generate_example_json(ServerConfig))))


async def _generate() -> dict[str, Any]:
    from ..server import global_subapp_pkgs

    cors_options = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }

    subapps: list[web.Application] = []
    for subapp in global_subapp_pkgs:
        pkg = importlib.import_module("ai.backend.appproxy.coordinator.api" + subapp)
        app, _ = pkg.create_app(cors_options)
        subapps.append(app)
    return generate_openapi("Proxy Coordinator", subapps, verbose=True)


@main.command()
@click.option(
    "--output",
    "-o",
    default="-",
    type=click.Path(dir_okay=False, writable=True),
    help="Output file path (default: stdout)",
)
def generate_openapi_spec(output: Path) -> None:
    """
    Generates OpenAPI specification of Backend.AI API.
    """
    openapi = asyncio.run(_generate())
    if output == "-" or output is None:
        print(json.dumps(openapi, ensure_ascii=False, indent=2))
    else:
        with open(output, mode="w") as fw:
            fw.write(json.dumps(openapi, ensure_ascii=False, indent=2))


@main.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--psql-container",
    "container_name",
    type=str,
    default=None,
    metavar="ID_OR_NAME",
    help=(
        "Open a postgres client shell using the psql executable "
        "shipped with the given postgres container. "
        'If not set or set as an empty string "", it will auto-detect '
        "the psql container from the halfstack. "
        'If set "-", it will use the host-provided psql executable. '
        "You may append additional arguments passed to the psql cli command. "
        "[default: auto-detect from halfstack]"
    ),
)
@click.option(
    "--psql-help",
    is_flag=True,
    help="Show the help text of the psql command instead of this dbshell command.",
)
@click.argument("psql_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_obj
def dbshell(cli_ctx: CLIContext, container_name, psql_help, psql_args):
    """
    Run the database shell.

    All additional arguments and options except `--psql-container` and `--psql-help` are
    transparently forwarded to the psql command.
    For instance, you can use `-c` to execute a psql/SQL statement on the command line.

    Note that you do not have to specify connection-related options
    because the dbshell command fills out them from the manager configuration.
    """
    db_config = cli_ctx.local_config.db
    if psql_help:
        psql_args = ["--help"]
    if not container_name:
        # Try to get the database container name of the halfstack
        candidate_container_names = subprocess.check_output(
            ["docker", "ps", "--format", "{{.Names}}", "--filter", "name=half-db"],
        )
        if not candidate_container_names:
            click.echo(
                "Could not find the halfstack postgres container. "
                "Please set the container name explicitly.",
                err=True,
            )
            sys.exit(ExitCode.FAILURE)
        container_name = None
        name_list = candidate_container_names.decode("utf-8").splitlines()
        for name in name_list:
            if "exporter" in name:
                continue  # Skip exporter containers
            container_name = name
            break
        if not container_name:
            click.echo(
                "Could not find the halfstack postgres container. "
                "Please set the container name explicitly.",
                err=True,
            )
            sys.exit(ExitCode.FAILURE)
    elif container_name == "-":
        # Use the host-provided psql command
        cmd = [
            "psql",
            (f"postgres://{db_config.user}:{db_config.password}@{db_config.addr}/{db_config.name}"),
            *psql_args,
        ]
        subprocess.run(cmd)
        return
    # Use the container to start the psql client command
    log.info(f"using the db container {container_name} ...")
    cmd = [
        "docker",
        "exec",
        "-i",
        "-t",
        container_name,
        "psql",
        "-U",
        db_config.user,
        "-d",
        db_config.name,
        *psql_args,
    ]
    subprocess.run(cmd)


if __name__ == "__main__":
    main()



import dataclasses
from typing import Annotated, Iterable

import aiohttp_cors
from aiohttp import web
from pydantic import BaseModel, Field

from ai.backend.appproxy.common.types import CORSOptions, PydanticResponse, WebMiddleware
from ai.backend.appproxy.common.utils import pydantic_api_handler

from ..models.worker import Worker
from ..types import RootContext
from .types import SlotModel
from .utils import auth_required


class ListSlotsRequestModel(BaseModel):
    wsproxy_host: Annotated[
        str | None,
        Field(
            default=None,
            description="Authority string (not UUID) of AppProxy worker. If not set, API will return slots of every workers.",
        ),
    ]
    in_use: Annotated[
        bool,
        Field(default=True, description="If set true, only returns information of occupied slots."),
    ]


class ListSlotsResponseModel(BaseModel):
    slots: list[SlotModel]


@auth_required("worker")
@pydantic_api_handler(ListSlotsRequestModel)
async def slots(
    request: web.Request, params: ListSlotsRequestModel
) -> PydanticResponse[ListSlotsResponseModel]:
    """
    Provides slot information hosted by worker mentioned.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async with root_ctx.db.begin_readonly_session() as sess:
        if _authority := params.wsproxy_host:
            workers = [await Worker.find_by_authority(sess, _authority, load_circuits=True)]
        else:
            workers = await Worker.list_workers(sess, load_circuits=True)
        slots: list[SlotModel] = []

        for worker in workers:
            slots += [
                SlotModel(**dataclasses.asdict(s))
                for s in (await worker.list_slots(sess))
                if s.in_use or (not params.in_use)
            ]

    return PydanticResponse(ListSlotsResponseModel(slots=slots))


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "api/slots"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("GET", slots))
    return app, []



import enum
import os
import socket
import sys
from pathlib import Path
from pprint import pformat
from typing import Annotated

import click
from pydantic import Field, FilePath, ValidationError

from ai.backend.appproxy.common.config import (
    BaseSchema,
    DebugConfig,
    GroupID,
    HostPortPair,
    PermitHashConfig,
    ProfilingConfig,
    RedisConfig,
    SecretConfig,
    UserID,
)
from ai.backend.appproxy.common.exceptions import ConfigValidationError
from ai.backend.appproxy.common.types import EventLoopType
from ai.backend.common import config
from ai.backend.common.types import ServiceDiscoveryType
from ai.backend.logging import LogLevel
from ai.backend.logging.config import LoggingConfig

_file_perm = (Path(__file__).parent / "server.py").stat()


class DBType(enum.StrEnum):
    POSTGRESQL = "postgresql"


class DistributedLock(enum.StrEnum):
    FILELOCK = "filelock"
    PG_ADVISORY = "pg_advisory"
    REDLOCK = "redlock"


class RedisLockConfig(BaseSchema):
    lock_retry_interval: Annotated[float | None, Field(default=None)]


class DBConfig(BaseSchema):
    type: Annotated[DBType, Field(description="Database type.", examples=[DBType.POSTGRESQL])]
    addr: Annotated[
        HostPortPair,
        Field(
            description="Address and port number of database server.",
            examples=[HostPortPair(host="127.0.0.1", port=8201)],
        ),
    ]
    name: Annotated[
        str, Field(min_length=2, max_length=64, description="Database name.", examples=["appproxy"])
    ]
    user: Annotated[str, Field(description="Database username.", examples=["backend"])]
    password: Annotated[str, Field(description="Databsase password.", examples=["develove"])]
    pool_size: Annotated[int, Field(gt=0, default=8)]
    max_overflow: Annotated[int, Field(gt=-2, default=64)]


class EtcdConfig(BaseSchema):
    addr: HostPortPair
    password: Annotated[str | None, Field(default=None)]
    namespace: Annotated[str, Field(default="traefik")]


class TraefikConfig(BaseSchema):
    etcd: EtcdConfig


class OTELConfig(BaseSchema):
    enabled: Annotated[
        bool,
        Field(
            default=False,
            description=(
                "Whether to enable OpenTelemetry for tracing or logging. "
                "When enabled, traces or log will be collected and sent to the configured OTLP endpoint."
            ),
            examples=[True, False],
        ),
    ]
    log_level: Annotated[
        str,
        Field(
            default="INFO",
            description="Log level for OpenTelemetry logging.",
            examples=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        ),
    ]
    endpoint: Annotated[
        str,
        Field(
            default="http://localhost:4317",
            description=(
                "OTLP (OpenTelemetry Protocol) endpoint for exporting telemetry data. "
                "Should include the host and port of the OTLP receiver."
            ),
            examples=["http://localhost:4317", "http://otel-collector:4317"],
        ),
    ]


class ServiceDiscoveryConfig(BaseSchema):
    type: Annotated[
        ServiceDiscoveryType,
        Field(
            default=ServiceDiscoveryType.REDIS,
            description="Type of service discovery to use.",
            examples=[item.value for item in ServiceDiscoveryType],
        ),
    ]


class ProxyCoordinatorConfig(BaseSchema):
    ipc_base_path: Annotated[
        Path,
        Field(
            default=Path("/tmp/backend.ai/ipc"),
            description="Directory to store temporary UNIX sockets.",
        ),
    ]
    event_loop: Annotated[
        EventLoopType,
        Field(default=EventLoopType.ASYNCIO, description="Type of event loop to use."),
    ]
    pid_file: Annotated[
        Path,
        Field(
            default=Path(os.devnull),
            description="Place to store process PID.",
            examples=["/run/backend.ai/appproxy/coordinator.pid"],
        ),
    ]

    id: Annotated[str, Field(default=f"i-{socket.gethostname()}", description="Node id.")]
    user: Annotated[
        int,
        UserID(default_uid=_file_perm.st_uid),
        Field(default=_file_perm.st_uid, description="Process owner."),
    ]
    group: Annotated[
        int,
        GroupID(default_gid=_file_perm.st_gid),
        Field(default=_file_perm.st_uid, description="Process group."),
    ]

    bind_addr: Annotated[
        HostPortPair,
        Field(
            default=HostPortPair(host="0.0.0.0", port=10200),
            description="Hostname and port number coordinator API server will listen to.",
        ),
    ]
    advertised_addr: Annotated[
        HostPortPair | None,
        Field(
            default=None,
            description="Hostname and port number coordinator API server will try to connect to.",
        ),
    ]

    distributed_lock: Annotated[
        DistributedLock,
        Field(
            default=DistributedLock.PG_ADVISORY,
            description="Connection backend of the distribution lock used in coordinator.",
        ),
    ]
    redlock_config: Annotated[
        RedisLockConfig, Field(default=RedisLockConfig(), description="Redis lock configuration")
    ]

    tls_listen: Annotated[bool, Field(default=False, description="Opt in to HTTPS setup.")]
    tls_cert: Annotated[
        FilePath | None,
        Field(
            default=None,
            description="Path to TLS certificate PEM.",
            examples=["/etc/backend.ai/tls/cert.pem"],
        ),
    ]
    tls_privkey: Annotated[
        FilePath | None,
        Field(
            default=None,
            description="Path to TLS private key PEM.",
            examples=["/etc/backend.ai/tls/privkey.pem"],
        ),
    ]
    tls_advertised: Annotated[
        bool,
        Field(
            default=False,
            description="Must be active if proxy coordinator is served behind external TLS terminator.",
        ),
    ]

    allow_unauthorized_configure_request: Annotated[
        bool,
        Field(
            default=False,
            description="Do not require access token on /v2/conf request. Must not be set true unless using outdated Backend.AI cluster.",
        ),
    ]

    use_experimental_redis_event_dispatcher: Annotated[
        bool,
        Field(
            default=False,
            description="Use experimental Redis event dispatcher implementation.",
        ),
    ]

    enable_traefik: Annotated[
        bool,
        Field(
            default=False,
            description="Use traefik as proxy worker's data plane.",
        ),
    ]
    traefik: Annotated[TraefikConfig | None, Field(default=None)]

    worker_heartbeat_timeout: Annotated[
        float,
        Field(
            default=30.0,
            description="Lifetime for each worker. At the time of proxy initiation request, workers which have heartbeat timestamp older than this threshold will be excluded from scheduling.",
        ),
    ]
    aiomonitor_termui_port: Annotated[
        int,
        Field(
            gt=0, lt=65536, description="Port number for aiomonitor termui server.", default=48500
        ),
    ]
    aiomonitor_webui_port: Annotated[
        int,
        Field(
            gt=0, lt=65536, description="Port number for aiomonitor webui server.", default=49500
        ),
    ]
    unused_circuit_collection_timeout: Annotated[
        int,
        Field(
            default=3600,
            description="Grace period for inactive circuits. Circuits who did not report any traffics for that amount of time will be shut. Not applicable for inference circuits.",
        ),
    ]

    metric_access_allowed_hosts: str = Field(
        default="127.0.0.1/32",
        description="Limits access to `/metrics` HTTP resources from described hosts only.",
    )

    health_check_timer_interval: Annotated[
        float,
        Field(
            default=30.0,
            description="Interval in seconds for model service health check timer",
        ),
    ]

    announce_addr: Annotated[
        HostPortPair,
        Field(
            default_factory=lambda: HostPortPair(host="127.0.0.1", port=10200),
            description=("Manually set the announced address for service discovery. "),
        ),
    ]

    @property
    def advertise_base_url(self) -> str:
        """
        Generate the full advertise URL with protocol.
        Uses advertised_addr if set, otherwise falls back to bind_addr.
        Protocol is determined by tls_advertised or tls_listen flags.
        """
        connection_info = (
            self.advertised_addr if self.advertised_addr is not None else self.bind_addr
        )
        if connection_info.host_set_with_protocol:
            return f"{connection_info.host}:{connection_info.port}"
        protocol = "https" if (self.tls_advertised or self.tls_listen) else "http"
        return f"{protocol}://{connection_info.host}:{connection_info.port}"


class ServerConfig(BaseSchema):
    db: DBConfig
    redis: RedisConfig
    core_redis: Annotated[
        RedisConfig | None,
        Field(
            default=None,
            description="Redis configuration for Backend.AI event publishing. Required if appproxy's primary redis backend differs from core's one.",
        ),
    ] = None
    proxy_coordinator: ProxyCoordinatorConfig
    profiling: Annotated[ProfilingConfig, Field(default_factory=lambda: ProfilingConfig())]
    secrets: SecretConfig
    permit_hash: PermitHashConfig
    logging: LoggingConfig
    debug: DebugConfig
    otel: Annotated[
        OTELConfig,
        Field(
            default_factory=lambda: OTELConfig(
                enabled=False, log_level="INFO", endpoint="http://localhost:4317"
            )
        ),
    ]
    service_discovery: Annotated[
        ServiceDiscoveryConfig,
        Field(default_factory=lambda: ServiceDiscoveryConfig(type=ServiceDiscoveryType.REDIS)),
    ]


def load(config_path: Path | None = None, log_level: LogLevel = LogLevel.NOTSET) -> ServerConfig:
    # Determine where to read configuration.
    raw_cfg, _ = config.read_from_file(config_path, "app-proxy-coordinator")

    config.override_key(raw_cfg, ("debug", "enabled"), log_level == LogLevel.DEBUG)
    if log_level != LogLevel.NOTSET:
        config.override_key(raw_cfg, ("logging", "level"), log_level)
        config.override_key(raw_cfg, ("logging", "pkg-ns", "ai.backend"), log_level)

    # Validate and fill configurations
    # (allow_extra will make configs to be forward-copmatible)
    try:
        server_config = ServerConfig(**raw_cfg)
        if server_config.profiling.enable_pyroscope:
            if not server_config.profiling.pyroscope_config:
                raise ConfigValidationError("Pyroscope enabled but config is not populated")
            if server_config.profiling.pyroscope_config.application_name is None:
                server_config.profiling.pyroscope_config.application_name = (
                    f"proxy-coordinator-{server_config.proxy_coordinator.bind_addr.port}"
                )
        if (
            server_config.proxy_coordinator.enable_traefik
            and not server_config.proxy_coordinator.traefik
        ):
            raise ConfigValidationError("Traefik enabled but configuration is not populated")
        if server_config.debug.enabled:
            print("== Proxy Coordinator configuration ==", file=sys.stderr)
            print(pformat(server_config.model_dump()), file=sys.stderr)
    except (ValidationError, ConfigValidationError) as e:
        print(
            "ConfigurationError: Could not read or validate the manager local config:",
            file=sys.stderr,
        )
        print(pformat(e), file=sys.stderr)
        raise click.Abort()
    else:
        return server_config



from __future__ import annotations

import textwrap
from datetime import datetime
from typing import TYPE_CHECKING, Annotated, Iterable
from uuid import UUID

import aiohttp_cors
import jwt
import sqlalchemy as sa
from aiohttp import web
from pydantic import AnyUrl, BaseModel, Field
from yarl import URL

from ai.backend.appproxy.common.exceptions import ObjectNotFound
from ai.backend.appproxy.common.types import (
    AppMode,
    CORSOptions,
    EndpointConfig,
    FrontendMode,
    HealthCheckConfig,
    ProxyProtocol,
    PydanticResponse,
    WebMiddleware,
)
from ai.backend.appproxy.common.utils import (
    pydantic_api_handler,
    pydantic_api_response_handler,
)
from ai.backend.appproxy.coordinator.models.worker import add_circuit

from ..models import Circuit, Endpoint, Worker
from ..models.utils import execute_with_txn_retry
from ..types import RootContext
from .types import SessionConfig, StubResponseModel
from .utils import auth_required

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession as SASession


class EndpointTagConfig(BaseModel):
    session: SessionConfig
    endpoint: EndpointConfig


class EndpointCreationRequestModel(BaseModel):
    version: Annotated[str, Field(description="Creation API version")]
    service_name: Annotated[str, Field(description="Name of the model service.")]
    tags: Annotated[
        EndpointTagConfig,
        Field(
            description="Metadata of target model service and dependent sessions.",
        ),
    ]
    open_to_public: Annotated[
        bool,
        Field(
            default=False,
            description=textwrap.dedent(
                """
                If set to true, AppProxy will require an API token (which can be obtained from `generate_endpoint_api_token` request)
                fullfilled at request header.
                """
            ),
        ),
    ] = False

    port: Annotated[int | None, Field(default=None, description="Preferred port number.")] = None
    subdomain: Annotated[
        str | None, Field(default=None, description="Preferred subdomain name.")
    ] = None

    health_check: Annotated[
        HealthCheckConfig | None,
        Field(
            default=None,
            description=textwrap.dedent(
                """
                Health check configuration for the model service.
                If provided, enables health checking for this endpoint.
                Matches model-definition.yaml health check schema.
                Only used when version is 'v2'.
                """
            ),
        ),
    ] = None


class EndpointCreationResponseModel(BaseModel):
    success: bool
    endpoint: AnyUrl
    health_check_enabled: bool


@auth_required("manager")
@pydantic_api_handler(EndpointCreationRequestModel)
async def create_or_update_endpoint(
    request: web.Request, params: EndpointCreationRequestModel
) -> PydanticResponse[EndpointCreationResponseModel]:
    """
    Creates or updates an inference circuit (v1) or syncs endpoint health check configuration (v2).
    Version is determined by the 'version' field in the request body.
    """
    root_ctx: RootContext = request.app["_root.context"]
    endpoint_id = UUID(request.match_info["endpoint_id"])

    health_check_config = params.health_check
    health_check_enabled = health_check_config is not None

    async def _sync(sess: SASession) -> URL:
        # Check if endpoint already exists
        try:
            endpoint = await Endpoint.get(sess, endpoint_id, load_circuit=False)
            # Update health check configuration
            endpoint.health_check_enabled = health_check_enabled
            endpoint.health_check_config = health_check_config
        except ObjectNotFound:
            # Create new endpoint record
            endpoint = Endpoint.create(
                endpoint_id=endpoint_id,
                health_check_enabled=health_check_enabled,
                health_check_config=health_check_config,
            )
            sess.add(endpoint)

        # Check if circuit already exists for this endpoint
        try:
            circuit = await Circuit.get_by_endpoint(
                sess, endpoint_id, load_worker=True, load_endpoint=True
            )
            circuit.endpoint_row = endpoint
            # Return existing circuit URL
            return await circuit.get_endpoint_url()
        except ObjectNotFound:
            pass  # Continue with creating new circuit

        # supported for subdomain based workers only
        matched_worker_id: UUID | None = None
        if _url := params.tags.endpoint.existing_url:
            assert _url.host
            domain = "." + ".".join(_url.host.split(".")[1:])

            query = sa.select(Worker).where(
                (
                    Worker.accepted_traffics.contains([AppMode.INFERENCE])
                    & (Worker.frontend_mode == FrontendMode.WILDCARD_DOMAIN)
                    & (Worker.wildcard_domain == domain)
                )
            )
            result = await sess.execute(query)
            matched_worker = result.scalar()
            if matched_worker:
                params.subdomain = _url.host.split(".")[0]
            else:
                assert _url.port
                query = sa.select(Worker).where(
                    (
                        Worker.accepted_traffics.contains([AppMode.INFERENCE])
                        & (Worker.frontend_mode == FrontendMode.PORT)
                        & (Worker.hostname == _url.host)
                    )
                )
                result = await sess.execute(query)
                worker_candidates = result.scalars().all()
                matched_workers = [
                    w
                    for w in worker_candidates
                    if w.port_range and w.port_range[0] <= _url.port <= w.port_range[1]
                ]
                if matched_workers:
                    params.port = _url.port
                    matched_worker = matched_workers[0]
                else:
                    matched_worker = None
            if matched_worker:
                matched_worker_id = matched_worker.id
        circuit, worker = await add_circuit(
            sess,
            params.tags.session,
            params.tags.endpoint,
            params.service_name,
            ProxyProtocol.HTTP,
            AppMode.INFERENCE,
            [],
            open_to_public=params.open_to_public,
            preferred_port=params.port,
            preferred_subdomain=params.subdomain or params.service_name,
            worker_id=matched_worker_id,
        )
        circuit.endpoint_id = endpoint.id
        circuit.endpoint_row = endpoint
        await root_ctx.circuit_manager.initialize_circuits([circuit])

        # Circuit already references endpoint by endpoint_id, no need to update
        # The relationship is handled automatically through endpoint_id matching

        # Route health status is now managed directly in the JSON route_info
        # Health status will be populated by the health checker when enabled

        await sess.flush()
        return await circuit.get_endpoint_url()

    async with root_ctx.db.connect() as db_conn:
        endpoint_url = await execute_with_txn_retry(_sync, root_ctx.db.begin_session, db_conn)

    return PydanticResponse(
        EndpointCreationResponseModel(
            success=True,
            endpoint=AnyUrl(str(endpoint_url)),
            health_check_enabled=health_check_enabled,
        )
    )


@auth_required("manager")
@pydantic_api_response_handler
async def remove_endpoint(request: web.Request) -> PydanticResponse[StubResponseModel]:
    """
    Deassociates inference circuit from system.
    """
    root_ctx: RootContext = request.app["_root.context"]

    endpoint_id = UUID(request.match_info["endpoint_id"])

    async def _update(sess: SASession) -> None:
        endpoint = await Endpoint.get(sess, endpoint_id, load_circuit=True)
        circuit = await Circuit.get(sess, endpoint.circuit_row.id, load_worker=True)
        circuit.worker_row.occupied_slots -= 1
        await sess.delete(circuit)
        await sess.delete(endpoint)

        await root_ctx.circuit_manager.unload_circuits([circuit])

    async with root_ctx.db.connect() as db_conn:
        await execute_with_txn_retry(_update, root_ctx.db.begin_session, db_conn)
    return PydanticResponse(StubResponseModel(success=True))


class UpdateHealthCheckConfigRequestModel(BaseModel):
    health_check: HealthCheckConfig | None


@auth_required("manager")
@pydantic_api_handler(UpdateHealthCheckConfigRequestModel)
async def inject_health_check_information(
    request: web.Request, params: UpdateHealthCheckConfigRequestModel
) -> PydanticResponse[StubResponseModel]:
    """
    Creates and returns API token required for execution of model service apps hosted by AppProxy.
     This API is meant to be called from Backend.AI manager rather than model service callee itself.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async with root_ctx.db.begin_session() as sess:
        circuit: Circuit = await Circuit.find_by_endpoint(
            sess, UUID(request.match_info["endpoint_id"]), load_worker=False
        )
        if circuit.app_mode != AppMode.INFERENCE:
            raise ObjectNotFound(object_name="inference-circuit")

        assert circuit.endpoint_row
        circuit.endpoint_row.health_check_enabled = params.health_check is not None
        circuit.endpoint_row.health_check_config = params.health_check

    return PydanticResponse(StubResponseModel(success=True))


class EndpointAPITokenGenerationRequestModel(BaseModel):
    user_uuid: UUID
    """
    Token requester's user UUID.
    """
    exp: datetime
    """
    Expiration date of token.
    """


class EndpointAPITokenResponseModel(BaseModel):
    token: str


@auth_required("manager")
@pydantic_api_handler(EndpointAPITokenGenerationRequestModel)
async def generate_endpoint_api_token(
    request: web.Request, params: EndpointAPITokenGenerationRequestModel
) -> PydanticResponse[EndpointAPITokenResponseModel]:
    """
    Creates and returns API token required for execution of model service apps hosted by AppProxy.
     This API is meant to be called from Backend.AI manager rather than model service callee itself.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async with root_ctx.db.begin_readonly_session() as sess:
        circuit: Circuit = await Circuit.find_by_endpoint(
            sess, UUID(request.match_info["endpoint_id"]), load_worker=False, load_endpoint=False
        )
        payload = dict(circuit.dump_model())
        payload["config"] = {}
        payload["app_url"] = str(await circuit.get_endpoint_url(session=sess))

    payload["user"] = str(params.user_uuid)
    payload["exp"] = params.exp
    encoded_jwt = jwt.encode(payload, root_ctx.local_config.secrets.jwt_secret, algorithm="HS256")
    return PydanticResponse(EndpointAPITokenResponseModel(token=encoded_jwt))


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "v2/endpoints"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    cors.add(app.router.add_resource(r""))
    cors.add(add_route("POST", "/{endpoint_id}", create_or_update_endpoint))
    cors.add(add_route("PUT", "/{endpoint_id}/health-check", inject_health_check_information))
    cors.add(add_route("DELETE", "/{endpoint_id}", remove_endpoint))
    cors.add(add_route("POST", "/{endpoint_id}/token", generate_endpoint_api_token))
    return app, []



import functools
from typing import Callable, Literal

import jwt
from aiohttp import web
from aiohttp.typedefs import Handler

from ai.backend.appproxy.common.exceptions import AuthorizationFailed
from ai.backend.appproxy.common.utils import set_handler_attr
from ai.backend.appproxy.coordinator.types import RootContext


def auth_required(scope: Literal["manager"] | Literal["worker"]) -> Callable[[Handler], Handler]:
    def wrap(handler: Handler) -> Handler:
        @functools.wraps(handler)
        async def wrapped(request: web.Request, *args, **kwargs):
            root_ctx: RootContext = request.app["_root.context"]
            permitted_token = root_ctx.local_config.secrets.api_secret
            permitted_header_values = (
                permitted_token,
                f"Bearer {permitted_token}",
                f"BackendAI {permitted_token}",
            )
            if _token := request.headers.get("X-BackendAI-Token"):
                if _token not in permitted_header_values:
                    raise AuthorizationFailed("Unauthorized access")
            elif _auth_header := request.headers.get("Authorization"):
                method, _, token = _auth_header.partition(" ")
                match method.lower():
                    case "bearer":
                        try:
                            jwt.decode(
                                token,
                                root_ctx.local_config.secrets.api_secret,
                                algorithms=["HS256"],
                            )
                        except jwt.PyJWTError:
                            raise AuthorizationFailed("Unauthorized access")
                    case _:
                        raise AuthorizationFailed(f"Unsupported authorization method {method}")
            else:
                raise AuthorizationFailed("Unauthorized access")
            return await handler(request, *args, **kwargs)

        original_attrs = getattr(handler, "_backend_attrs", {})
        for k, v in original_attrs.items():
            set_handler_attr(wrapped, k, v)

        set_handler_attr(wrapped, "auth_scope", scope)
        return wrapped

    return wrap



from __future__ import annotations

import dataclasses
import logging
import textwrap
import uuid
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Annotated, Iterable
from uuid import UUID

import aiohttp_cors
import attrs
from aiohttp import web
from dateutil.tz import tzutc
from pydantic import BaseModel, Field

from ai.backend.appproxy.common.config import get_default_redis_key_ttl
from ai.backend.appproxy.common.events import DoCheckWorkerLostEvent, WorkerLostEvent
from ai.backend.appproxy.common.exceptions import ObjectNotFound
from ai.backend.appproxy.common.types import (
    AppMode,
    CORSOptions,
    FrontendMode,
    ProxyProtocol,
    PydanticResponse,
    SerializableCircuit,
    WebMiddleware,
)
from ai.backend.appproxy.common.utils import (
    pydantic_api_handler,
    pydantic_api_response_handler,
)
from ai.backend.appproxy.coordinator.defs import LockID
from ai.backend.common.distributed import GlobalTimer
from ai.backend.common.events.dispatcher import EventHandler
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter

from ..models import Token, Worker, WorkerAppFilter, WorkerStatus
from ..models.utils import execute_with_txn_retry
from ..types import RootContext
from .types import CircuitListResponseModel, SlotModel, StubResponseModel
from .utils import auth_required

if TYPE_CHECKING:
    pass
    from sqlalchemy.ext.asyncio import AsyncSession as SASession

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]


class WorkerModel(BaseModel):
    authority: Annotated[
        str,
        Field(
            description="authority string of worker. Unique across every workers joined on a single coordinator."
        ),
    ]
    frontend_mode: FrontendMode
    protocol: ProxyProtocol

    hostname: str
    tls_listen: bool
    tls_advertised: bool

    api_port: int

    port_range: Annotated[tuple[int, int] | None, Field(default=None)]
    wildcard_domain: Annotated[str | None, Field(default=None)]
    wildcard_traffic_port: Annotated[int | None, Field(default=None)]
    filtered_apps_only: bool
    traefik_last_used_marker_path: Annotated[str | None, Field(default=None)]

    accepted_traffics: list[AppMode]


class AppFilter(BaseModel):
    key: str
    value: str


class WorkerRequestModel(WorkerModel):
    app_filters: Annotated[list[AppFilter], Field(default=[])]


class WorkerResponseModel(WorkerModel):
    id: Annotated[UUID, Field(description="ID of worker.")]

    created_at: datetime
    updated_at: datetime

    available_slots: Annotated[
        int,
        Field(
            description=textwrap.dedent(
                """
                Number of slots worker is capable to hold. Workers serving `subdomain` frontend have -1 as `available_circuits`.
                For `port` frontend this value is number of ports exposed by the worker.
                """
            ),
        ),
    ]

    occupied_slots: Annotated[int, Field(description="Number of slots occupied by circuit.")]

    nodes: Annotated[
        int,
        Field(
            description="Number of actual nodes claiming as same worker. Can be considered as HA set up if this value is greater than 1.",
        ),
    ]

    slots: list[SlotModel]


class WorkerListResponseModel(BaseModel):
    workers: list[WorkerResponseModel]


class TokenResponseModel(BaseModel):
    login_session_token: str | None
    kernel_host: str
    kernel_port: int
    session_id: UUID
    user_uuid: UUID
    group_id: UUID
    access_key: str
    domain_name: str


@auth_required("worker")
@pydantic_api_response_handler
async def get_worker(request: web.Request) -> PydanticResponse[WorkerResponseModel]:
    """
    Returns information about worker mentioned.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async with root_ctx.db.begin_readonly_session() as sess:
        worker = await Worker.get(sess, UUID(request.match_info["worker_id"]))
        return PydanticResponse(
            WorkerResponseModel(
                slots=[SlotModel(**dataclasses.asdict(s)) for s in await worker.list_slots(sess)],
                **worker.dump_model(),
            )
        )


@auth_required("worker")
@pydantic_api_response_handler
async def list_workers(request: web.Request) -> PydanticResponse[WorkerListResponseModel]:
    """
    Lists all workers recognized by coordinator.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async with root_ctx.db.begin_readonly_session() as sess:
        workers = await Worker.list_workers(sess)
        return PydanticResponse(
            WorkerListResponseModel(
                workers=[
                    WorkerResponseModel(
                        slots=[
                            SlotModel(**dataclasses.asdict(s)) for s in await w.list_slots(sess)
                        ],
                        **w.dump_model(),
                    )
                    for w in workers
                ]
            )
        )


@auth_required("worker")
@pydantic_api_response_handler
async def list_worker_circuits(request: web.Request) -> PydanticResponse[CircuitListResponseModel]:
    """
    Lists every circuits worker is currently serving.
    """
    root_ctx: RootContext = request.app["_root.context"]

    async with root_ctx.db.begin_readonly_session() as sess:
        worker = await Worker.get(sess, UUID(request.match_info["worker_id"]), load_circuits=True)
        return PydanticResponse(
            CircuitListResponseModel(
                circuits=[SerializableCircuit(**c.dump_model()) for c in worker.circuits]
            )
        )


@auth_required("worker")
@pydantic_api_handler(WorkerRequestModel)
async def update_worker(
    request: web.Request, params: WorkerRequestModel
) -> PydanticResponse[WorkerResponseModel]:
    """
    Registers worker to coordinator.
    """

    root_ctx: RootContext = request.app["_root.context"]

    async def _update(sess: SASession) -> dict:
        try:
            worker = await Worker.find_by_authority(sess, params.authority)
            worker.frontend_mode = params.frontend_mode
            worker.protocol = params.protocol
            worker.hostname = params.hostname
            worker.tls_listen = params.tls_listen
            worker.tls_advertised = params.tls_advertised
            worker.api_port = params.api_port
            worker.port_range = params.port_range
            worker.wildcard_domain = params.wildcard_domain
            worker.wildcard_traffic_port = params.wildcard_traffic_port
            worker.filtered_apps_only = params.filtered_apps_only
            worker.traefik_last_used_marker_path = params.traefik_last_used_marker_path
            worker.updated_at = datetime.now()
            worker.nodes += 1
            worker.status = WorkerStatus.ALIVE
        except ObjectNotFound:
            worker = Worker.create(
                uuid.uuid4(),
                params.authority,
                params.frontend_mode,
                params.protocol,
                params.hostname,
                params.tls_listen,
                params.tls_advertised,
                params.api_port,
                params.accepted_traffics,
                port_range=params.port_range,
                wildcard_domain=params.wildcard_domain,
                wildcard_traffic_port=params.wildcard_traffic_port,
                filtered_apps_only=params.filtered_apps_only,
                traefik_last_used_marker_path=params.traefik_last_used_marker_path,
                status=WorkerStatus.ALIVE,
            )
            sess.add(worker)
            await sess.flush()
            await sess.refresh(worker)

        for filter in params.app_filters:
            try:
                await WorkerAppFilter.find_by_rule(sess, worker.id, filter.key, filter.value)
            except ObjectNotFound:
                filter_row = WorkerAppFilter.create(
                    uuid.uuid4(),
                    property_name=filter.key,
                    property_value=filter.value,
                    worker=worker.id,
                )
                sess.add(filter_row)

        result = dict(worker.dump_model())
        result["slots"] = [
            SlotModel(**dataclasses.asdict(s)) for s in (await worker.list_slots(sess))
        ]
        log.info("Worker {} joined", worker.authority)
        return result

    async with root_ctx.db.connect() as db_conn:
        result = await execute_with_txn_retry(_update, root_ctx.db.begin_session, db_conn)
    return PydanticResponse(WorkerResponseModel(**result))


@auth_required("worker")
@pydantic_api_response_handler
async def delete_worker(request: web.Request) -> PydanticResponse[StubResponseModel]:
    """
    Deassociates worker from coordinator.
    """
    root_ctx: RootContext = request.app["_root.context"]
    worker_id = UUID(request.match_info["worker_id"])

    async def _update(sess: SASession) -> None:
        worker = await Worker.get(sess, worker_id)
        worker.nodes -= 1
        if worker.nodes == 0:
            worker.status = WorkerStatus.LOST

    async with root_ctx.db.connect() as db_conn:
        await execute_with_txn_retry(_update, root_ctx.db.begin_session, db_conn)
    return PydanticResponse(StubResponseModel(success=True))


@auth_required("worker")
@pydantic_api_response_handler
async def heartbeat_worker(request: web.Request) -> PydanticResponse[WorkerResponseModel]:
    root_ctx: RootContext = request.app["_root.context"]
    worker_id = UUID(request.match_info["worker_id"])
    now = datetime.now(tzutc())

    async def _update(sess: SASession) -> dict:
        worker = await Worker.get(sess, worker_id)
        worker.updated_at = datetime.now()
        worker.status = WorkerStatus.ALIVE
        result = dict(worker.dump_model())
        result["slots"] = [
            SlotModel(**dataclasses.asdict(s)) for s in (await worker.list_slots(sess))
        ]

        # Update "last seen" timestamp for liveness tracking
        ttl = get_default_redis_key_ttl()
        await root_ctx.valkey_live.hset_with_expiry(
            "proxy-worker.last_seen",
            {worker.authority: str(now.timestamp())},
            ttl,
        )
        return result

    async with root_ctx.db.connect() as db_conn:
        result = await execute_with_txn_retry(_update, root_ctx.db.begin_session, db_conn)
    request["do_not_print_access_log"] = True
    return PydanticResponse(WorkerResponseModel(**result))


@auth_required("worker")
@pydantic_api_response_handler
async def get_token(request: web.Request) -> PydanticResponse[TokenResponseModel]:
    """
    Lists tokens issued within /v2/conf request handler.
    """
    root_ctx: RootContext = request.app["_root.context"]
    token_id = UUID(request.match_info["token_id"])

    async with root_ctx.db.begin_readonly_session() as sess:
        token = await Token.get(sess, token_id)
        return PydanticResponse(TokenResponseModel(**token.dump_model()))


async def check_worker_lost(
    app: web.Application, src: AgentId, event: DoCheckWorkerLostEvent
) -> None:
    root_ctx: RootContext = app["_root.context"]

    try:
        now = datetime.now(tzutc())
        timeout = timedelta(
            seconds=root_ctx.local_config.proxy_coordinator.worker_heartbeat_timeout
        )

        msg_data = await root_ctx.valkey_live.hgetall_str("proxy-worker.last_seen")

        async with root_ctx.db.begin_readonly_session() as sess:
            workers = await Worker.list_workers(sess)
            worker_map = {w.authority: w for w in workers}

        for worker_id_str, prev in msg_data.items():
            prev = datetime.fromtimestamp(float(prev), tzutc())
            if (
                (now - prev) > timeout
                and worker_id_str in worker_map
                and worker_map[worker_id_str].status == WorkerStatus.ALIVE
            ):
                await root_ctx.event_producer.anycast_event(
                    WorkerLostEvent(worker_id_str, "heartbeat timeout")
                )
    except Exception:
        log.exception("check_worker_lost(): exception:")


@attrs.define(slots=True, auto_attribs=True, init=False)
class PrivateContext:
    worker_lost_check_timer: GlobalTimer
    worker_lost_check_evh: EventHandler[web.Application, DoCheckWorkerLostEvent]


async def init(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["worker.context"]

    # Scan ALIVE workers
    app_ctx.worker_lost_check_evh = root_ctx.event_dispatcher.consume(
        DoCheckWorkerLostEvent,
        app,
        check_worker_lost,
    )
    app_ctx.worker_lost_check_timer = GlobalTimer(
        root_ctx.distributed_lock_factory(LockID.LOCKID_WORKER_LOST, 15.0),
        root_ctx.event_producer,
        lambda: DoCheckWorkerLostEvent(),
        15.0,
        initial_delay=10.0,
        task_name="check_worker_lost_task",
    )
    await app_ctx.worker_lost_check_timer.join()


async def shutdown(app: web.Application) -> None:
    root_ctx: RootContext = app["_root.context"]
    app_ctx: PrivateContext = app["worker.context"]
    await app_ctx.worker_lost_check_timer.leave()
    root_ctx.event_dispatcher.unconsume(app_ctx.worker_lost_check_evh)


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "api/worker"
    app["worker.context"] = PrivateContext()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    root_resource = cors.add(app.router.add_resource(r""))
    cors.add(root_resource.add_route("GET", list_workers))
    cors.add(root_resource.add_route("PUT", update_worker))
    cors.add(add_route("GET", "/{worker_id}", get_worker))
    cors.add(add_route("PATCH", "/{worker_id}", heartbeat_worker))
    cors.add(add_route("GET", "/{worker_id}/circuits", list_worker_circuits))
    cors.add(add_route("DELETE", "/{worker_id}", delete_worker))
    return app, []
